from concurrent import futures
import threading
import logging
import queue

import grpc

from src.common.grpc.idl.connector_pb2_grpc import (
    ConnectorServicer,
    add_ConnectorServicer_to_server,
)
from src.common.grpc.idl.connector_pb2 import ConnectStreamResponse
from src.common.grpc.interceptors.authentication_server import AuthenticationInterceptor


class Connector(ConnectorServicer):
    def __init__(self, in_q: queue.Queue, out_q: queue.Queue) -> None:
        super().__init__()
        self.in_q = in_q
        self.out_q = out_q
        self.is_active = False
        self.recv_thread = None

    def ConnectStream(self, request_iterator, context):
        def recv_to_queue(request_iterator):
            logging.info("ConnectStream api recv_to_queue thread start")
            try:
                for request in request_iterator:
                    if request.frame is None:
                        logging.debug("connectStream request frame is None, break")
                        self.is_active = False
                        break
                    logging.debug(f"be in_q put ----> len(request.frame):{len(request.frame)}")
                    self.in_q.put(request.frame)
            except Exception as e:
                logging.error(f"ConnectStream api recv_to_queue error:{e}")

        # if self.recv_thread is None or self.recv_thread.is_alive() is False:
        self.recv_thread = threading.Thread(target=recv_to_queue, args=(request_iterator,))
        self.recv_thread.start()
        self.is_active = True

        logging.info("ConnectStream api wait out_q frame to yield")
        while self.is_active:
            frame = self.out_q.get()
            if frame is None:
                logging.debug("connectStream response frame is None, break")
                break
            logging.debug(f"be out_q put ----> len(frame):{len(frame)}")
            yield ConnectStreamResponse(frame=frame)

    def close(self):
        self.out_q.put(None)
        self.is_active = False
        if self.recv_thread is not None and self.recv_thread.is_alive():
            self.recv_thread.join()


class StreamServe:
    def __init__(
        self,
        in_q: queue.Queue,
        out_q: queue.Queue,
        token="chat-bot-connenctor",
        port="50052",
        max_workers=10,
    ) -> None:
        self.token = token
        self.port = port
        self.max_workers = max_workers
        self.connector = Connector(in_q, out_q)
        self.serve_thread: threading.Thread = None

        logging.info(f"serve port: {self.port} max_workers: {self.max_workers}")
        authenticator = AuthenticationInterceptor(
            "authorization", self.token, grpc.StatusCode.UNAUTHENTICATED, "Access denied!"
        )
        server = grpc.server(
            futures.ThreadPoolExecutor(max_workers=self.max_workers),
            interceptors=(authenticator,),
        )
        add_ConnectorServicer_to_server(self.connector, server)
        server.add_insecure_port(f"[::]:{self.port}")
        self.server = server

    def start(self):
        if self.serve_thread is None or self.serve_thread.is_alive() is False:
            self.serve_thread = threading.Thread(target=self._serve, args=())
            self.serve_thread.start()

    def stop(self):
        logging.debug("Server stoping...")
        self.server.stop(1).wait()
        logging.info("Server stop")
        if self.serve_thread is not None or self.serve_thread.is_alive():
            self.serve_thread.join()
            logging.info("serve_thread stop")

    def close(self):
        self.connector.close()
        self.stop()

    def _serve(self) -> None:
        self.server.start()
        logging.info(f"Server started port: {self.port}")
        self.server.wait_for_termination()



import copy
import os
import json
import logging
import asyncio
import time
from urllib.parse import quote
from typing import Any, AsyncIterator, Awaitable, Callable, List
import uuid

import PIL.Image
import numpy as np
from pydantic import BaseModel
from apipeline.frames.data_frames import AudioRawFrame, ImageRawFrame

from src.common import const
from src.common.types import SAMPLE_WIDTH, AgoraParams, LOG_DIR, VIDEOS_DIR
from src.common.utils.audio_utils import resample_audio
from src.types.frames.data_frames import (
    AgoraTransportMessageFrame,
    TransportMessageFrame,
    UserAudioRawFrame,
    UserImageRawFrame,
)
from src.services.help.agora.token import TokenClaims, TokenPaser
from src.services.help.agora.video_utils import *

try:
    from agora_realtime_ai_api import rtc
    from agora.rtc.voice_detection import AudioVadV2, AudioVadConfigV2
    from agora.rtc.video_frame_observer import VideoFrame
    from agora.rtc.video_frame_observer import IVideoFrameObserver
    from agora.rtc.video_frame_sender import ExternalVideoFrame
    from agora.rtc.agora_base import (
        VideoEncoderConfiguration,
        VideoDimensions,
        VideoSubscriptionOptions,
    )
except ModuleNotFoundError as e:
    logging.error(f"Exception: {e}")
    logging.error("In order to use Agora, you need to `pip install achatbot[agora]`.")
    raise Exception(f"Missing module: {e}")


class AgoraService:
    @staticmethod
    def init(params: AgoraParams) -> rtc.AgoraService:
        config = rtc.AgoraServiceConfig()
        # https://api-ref.agora.io/en/voice-sdk/react-native/4.x/API/enum_audioscenariotype.html
        config.audio_scenario = rtc.AudioScenarioType.AUDIO_SCENARIO_CHORUS  # default
        # if params.camera_in_enabled:
        #    config.audio_scenario = rtc.AudioScenarioType.AUDIO_SCENARIO_CHATROOM
        config.appid = params.app_id
        config.enable_video = int(params.camera_in_enabled)
        config.log_path = os.path.join(LOG_DIR, "agorasdk.log")

        agora_service = rtc.AgoraService()
        agora_service.initialize(config)

        return agora_service


class VideoStream:
    def __init__(self) -> None:
        self.queue: asyncio.Queue = asyncio.Queue()

    def __aiter__(self) -> AsyncIterator[VideoFrame]:
        return self

    async def __anext__(self) -> VideoFrame:
        item = await self.queue.get()
        if item is None:
            raise StopAsyncIteration

        return item


class RtcChannelEventObserver(IVideoFrameObserver, rtc.ChannelEventObserver):
    """
    if rtc.ChannelEventObserver no register event name,
    - need add for custom event to run
    - or override event name to run

    add IVideoFrameObserver to support sub video frame
    """

    def __init__(self, event_emitter: rtc.AsyncIOEventEmitter, options: rtc.RtcOptions) -> None:
        super().__init__(event_emitter, options)
        self.video_streams = dict[str, VideoStream]()

    def on_connection_failure(
        self, agora_rtc_conn: rtc.RTCConnection, conn_info: rtc.RTCConnInfo, reason: int
    ):
        super().on_connection_failure(agora_rtc_conn, conn_info, reason)
        # logging.error(f"Connected to RTC: {agora_rtc_conn} {conn_info} {reason}")
        self.emit_event("connection_failure", agora_rtc_conn, conn_info, reason)

    def on_connected(
        self, agora_rtc_conn: rtc.RTCConnection, conn_info: rtc.RTCConnInfo, reason: int
    ):
        super().on_connected(agora_rtc_conn, conn_info, reason)
        # logging.info(f"Connected to RTC: {agora_rtc_conn} {conn_info} {reason}")
        self.emit_event("connected", agora_rtc_conn, conn_info, reason)

    def on_disconnected(
        self, agora_rtc_conn: rtc.RTCConnection, conn_info: rtc.RTCConnInfo, reason: int
    ):
        super().on_disconnected(agora_rtc_conn, conn_info, reason)
        # logging.info(f"Disconnected to RTC: {agora_rtc_conn} {conn_info} {reason}")
        self.emit_event("disconnected", agora_rtc_conn, conn_info, reason)

    def on_playback_audio_frame_before_mixing(
        self,
        agora_local_user: rtc.LocalUser,
        channelId,
        uid,
        frame: rtc.AudioFrame,
        vad_result_state: int,
        vad_result_bytearray: bytearray,  # TODO: vad
    ):
        logging.debug(f"push AudioStream from channel:{channelId} uid:{uid} rtc.AudioFrame:{frame}")

        # fix: when use had joined, the audio_streams no uid audio stream
        if self.audio_streams.get(uid) is None:
            return 0

        return super().on_playback_audio_frame_before_mixing(
            agora_local_user, channelId, uid, frame
        )

    def on_video_subscribe_state_changed(
        self, agora_local_user, channel, user_id, old_state, new_state, elapse_since_last_state
    ):
        self.emit_event(
            "video_subscribe_state_changed",
            agora_local_user,
            channel,
            user_id,
            old_state,
            new_state,
            elapse_since_last_state,
        )

    def on_frame(self, channel_id, remote_uid, video_frame: VideoFrame):
        # on_video_frame, channel_id=room-bot, remote_uid=1867636435, type=1, width=640,
        # height=480, y_stride=640, u_stride=320, v_stride=320, len_y=307200,
        # len_u=76800, len_v=76800, len_alpha_buffer=0
        if remote_uid == "0":  # TODO: check
            logging.debug(
                f"on_video_frame, channel_id={channel_id},"
                f"remote_uid={remote_uid}, type={video_frame.type}, width={video_frame.width},"
                f"height={video_frame.height}, y_stride={video_frame.y_stride},"
                f"u_stride={video_frame.u_stride}, v_stride={video_frame.v_stride},"
                f"len_y={len(video_frame.y_buffer)}, len_u={len(video_frame.u_buffer)},"
                f"len_v={len(video_frame.v_buffer)},"
                f"len_alpha_buffer={len(video_frame.alpha_buffer) if video_frame.alpha_buffer else 0}"
            )
            return 1

        # fix: when use had joined, the video_streams no uid video stream
        if self.video_streams.get(remote_uid) is None:
            return 1

        self.loop.call_soon_threadsafe(self.video_streams[remote_uid].queue.put_nowait, video_frame)
        return 1


class RtcChannel(rtc.Channel):
    """
    rtc channel now support audio/video now
    """

    def __init__(
        self,
        params: AgoraParams,
        token_claims: TokenClaims,
        options: rtc.RtcOptions,
        loop: asyncio.AbstractEventLoop,
        service: rtc.AgoraService = None,
        vad_conf: AudioVadConfigV2 = None,  # TODO:vad
    ) -> None:
        self._param = params
        self._token_claims = token_claims
        self.token = token_claims.token

        self.options = options
        self.uid = options.uid
        self.channelId = options.channel_name
        self.enable_pcm_dump = options.enable_pcm_dump

        self.loop = loop

        self._service = service if service else AgoraService.init(self._param)

        # chat message
        self.chat = rtc.Chat(self)

        # connect
        self.connection_state = 0
        self.remote_users = dict[int, Any]()
        conn_config = rtc.RTCConnConfig(
            client_role_type=rtc.ClientRoleType.CLIENT_ROLE_BROADCASTER,
            channel_profile=rtc.ChannelProfileType.CHANNEL_PROFILE_LIVE_BROADCASTING,
        )
        # NOTE: use subscribe_audio don't to set auto_subscribe_audio
        # if params.audio_in_enabled:
        #    conn_config.auto_subscribe_audio = 1
        # NOTE: use subscribe_video don't to set auto_subscribe_video
        # if params.camera_in_enabled:
        #     conn_config.auto_subscribe_video = 1
        self.connection = self._service.create_rtc_connection(conn_config)

        # Create the event emitter for observer
        self.emitter = rtc.AsyncIOEventEmitter(loop)
        # Channel Event Observer, register event name
        self.channel_event_observer = RtcChannelEventObserver(
            self.emitter,
            options=options,
        )

        # connect regist event observer
        self.connection.register_observer(self.channel_event_observer)
        # local user
        self.local_user = self.connection.get_local_user()

        # register local user event observer
        self.local_user.register_local_user_observer(self.channel_event_observer)
        # 1.1 in audio for local user
        if params.audio_in_enabled:
            self.local_user.set_playback_audio_frame_before_mixing_parameters(
                options.channels, options.sample_rate
            )
            # enable_vad = int(params.vad_enabled)  # TODO: vad
            enable_vad = 0
            ret = self.local_user.register_audio_frame_observer(
                self.channel_event_observer, enable_vad, vad_conf
            )

            if ret < 0:
                raise Exception("register_audio_frame_observer failed")
            # self.local_user.subscribe_all_audio()

        # 1.2 in video for local user
        if params.camera_in_enabled:
            ret = self.local_user.register_video_frame_observer(self.channel_event_observer)
            if ret < 0:
                raise Exception("register_video_frame_observer failed")

        # 2. media node factory for out audio/video track
        self.media_node_factory = self._service.create_media_node_factory()
        # 2.1 out audio track
        if params.audio_out_enabled:
            self.audio_pcm_data_sender = self.media_node_factory.create_audio_pcm_data_sender()
            self.audio_track = self._service.create_custom_audio_track_pcm(
                self.audio_pcm_data_sender
            )

        # 2.2 out video track
        if params.camera_out_enabled:
            self.video_yuv_data_sender = self.media_node_factory.create_video_frame_sender()
            self.video_track = self._service.create_custom_video_track_frame(
                self.video_yuv_data_sender
            )
            video_config = VideoEncoderConfiguration(
                frame_rate=params.camera_out_framerate,
                dimensions=VideoDimensions(
                    width=params.camera_out_width, height=params.camera_out_height
                ),
                encode_alpha=0,
            )
            self.video_track.set_video_encoder_configuration(video_config)

        # 3. create a data stream for send stream message
        self.stream_id = self.connection.create_data_stream(False, False)
        self.stream_message_queue = asyncio.Queue()

        # create a task to process stream message
        def log_exception(t: asyncio.Task[Any]) -> None:
            if not t.cancelled() and t.exception():
                logging.error("unhandled exception", exc_info=t.exception())

        asyncio.create_task(self._process_stream_message()).add_done_callback(log_exception)

        # 4. emitter event callbacks (event can register multiple callbacks)
        # NOTE:register emit event name see rtc.ChannelEventObserver,
        # c/c++ rtc core lib(defined callback impl) to call and register callback
        self.on("user_joined", self.on_user_joined)
        self.on("connection_state_changed", self.on_connection_state_changed)
        self.on("audio_subscribe_state_changed", self.on_audio_subscribe_state_changed)
        self.on("video_subscribe_state_changed", self.on_video_subscribe_state_changed)
        self.on("user_left", self.on_user_left)

    def set_local_user_audio_track(self) -> None:
        if self._param.audio_out_enabled:
            self.audio_track.set_enabled(1)
            self.local_user.publish_audio(self.audio_track)

    def set_local_user_video_track(self) -> None:
        if self._param.camera_out_enabled:
            self.video_track.set_enabled(1)
            self.local_user.publish_video(self.video_track)

    async def clear_sender_video_buffer(self) -> None:
        """
        Clears the video buffer which is used to send.
        """
        self.video_track.clear_sender_buffer()

    async def subscribe_video(self, uid: int) -> None:
        """
        Subscribes to the video of a user.

        Parameters:
            uid: The user ID to subscribe to.
        """
        future = asyncio.Future()

        def callback(
            agora_local_user,
            channel,
            user_id,
            old_state,
            new_state,
            elapse_since_last_state,
        ):
            if new_state == 3:  # Successfully subscribed
                future.set_result(None)
            # elif new_state == 1:  # Subscription failed
            #     future.set_exception(
            #         Exception(
            #             f"Failed to subscribe {user_id} video: state changed from {old_state} to {new_state}"
            #         )
            #     )

        self.on("video_subscribe_state_changed", callback)
        self.local_user.subscribe_video(uid, VideoSubscriptionOptions())

        try:
            await future
        except Exception as e:
            raise Exception(f"video subscription failed for user {uid}: {str(e)}") from e
        finally:
            self.off("video_subscribe_state_changed", callback)

    async def unsubscribe_video(self, uid: int) -> None:
        """
        Unsubscribes from the video of a user.

        Parameters:
            uid: The user ID to unsubscribe from.
        """
        future = asyncio.Future()

        def callback(
            agora_local_user,
            channel,
            user_id,
            old_state,
            new_state,
            elapse_since_last_state,
        ):
            if new_state == 3:  # Successfully unsubscribed
                future.set_result(None)
            else:  # Failed to unsubscribe
                future.set_exception(
                    Exception(
                        f"Failed to unsubscribe {user_id} video: state changed from {old_state} to {new_state}"
                    )
                )

        self.on("video_subscribe_state_changed", callback)
        self.local_user.unsubscribe_video(uid)

        try:
            await future
        except Exception as e:
            raise Exception(f"video unsubscription failed for user {uid}: {str(e)}") from e
        finally:
            self.off("video_subscribe_state_changed", callback)

    def release(self):
        self.local_user.unregister_local_user_observer()
        if self._param.audio_in_enabled:
            self.local_user.unregister_audio_frame_observer()
        if self._param.audio_out_enabled:
            self.audio_track.set_enabled(0)
            self.local_user.unpublish_audio(self.audio_track)
            self.audio_pcm_data_sender.release()
            self.audio_pcm_data_sender = None
            self.audio_track.release()
            self.audio_track = None
        if self._param.camera_in_enabled:
            self.local_user.unregister_video_frame_observer()
        if self._param.camera_out_enabled:
            self.video_track.set_enabled(0)
            self.local_user.unpublish_video(self.video_track)
            self.video_yuv_data_sender.release()
            self.video_yuv_data_sender = None
            self.video_track.release()
            self.video_track = None

        self.media_node_factory.release()
        self.media_node_factory = None
        self.local_user.release()
        self.local_user = None
        self.channel_event_observer = None
        logging.info("local user's channel released")

    # Event Callback

    def on_connection_state_changed(self, agora_rtc_conn, conn_info, reason):
        # logging.debug(f"conn state {conn_info.state}")
        self.connection_state = conn_info.state

    def on_user_joined(self, agora_rtc_conn, user_id):
        # logging.debug(f"User {user_id} joined")
        self.remote_users[user_id] = True

    def on_user_left(self, agora_rtc_conn, user_id, reason):
        logging.info(f"User({type(user_id)}) {user_id} left reason: {reason}")

    def destory_when_user_left(self, user_id):
        # NOTE: like class obj destory in C++,
        # this will be called when the last emit event (user left) is triggered
        if user_id in self.remote_users:
            self.remote_users.pop(user_id, None)
        if user_id in self.channel_event_observer.audio_streams:
            audio_stream = self.channel_event_observer.audio_streams.pop(user_id, None)
            audio_stream.queue.put_nowait(None)
        if user_id in self.channel_event_observer.video_streams:
            video_stream = self.channel_event_observer.video_streams.pop(user_id, None)
            video_stream.queue.put_nowait(None)

    def on_audio_subscribe_state_changed(
        self,
        agora_local_user: rtc.LocalUser,
        channel: str,
        user_id: str,
        old_state: int,
        new_state: int,
        elapse_since_last_state: int,
    ):
        # logging.debug(f"{agora_local_user} {channel} {user_id} {old_state} {new_state}")
        if new_state == 3:  # Successfully subscribed
            if user_id not in self.channel_event_observer.audio_streams:
                self.channel_event_observer.audio_streams[user_id] = rtc.AudioStream()
                logging.info(
                    f"{channel} {user_id} new AudioStream in audio_streams:{self.channel_event_observer.audio_streams}"
                )

    def on_video_subscribe_state_changed(
        self,
        agora_local_user: rtc.LocalUser,
        channel: str,
        user_id: str,
        old_state: int,
        new_state: int,
        elapse_since_last_state: int,
    ):
        if new_state == 3:  # Successfully subscribed
            if user_id not in self.channel_event_observer.video_streams:
                self.channel_event_observer.video_streams[user_id] = VideoStream()
                logging.info(
                    f"{channel} {user_id} new VideoStream in video_streams:{self.channel_event_observer.video_streams}"
                )

    # Audio out
    async def push_pcm_audio_frame(self, pcm_audio_frame: rtc.PcmAudioFrame) -> None:
        """
        Pushes an audio frame to the channel.

        Parameters:
            frame: The pcm audio frame to be pushed.
        """
        frame = copy.copy(pcm_audio_frame)
        ret = self.audio_pcm_data_sender.send_audio_pcm_data(frame)
        logging.debug(
            f"Pushed audio frame: {ret}, audio frame length: {len(pcm_audio_frame.data) if pcm_audio_frame.data else 0}"
        )
        if ret and ret < 0:
            raise Exception(
                f"Failed to send audio frame: {ret}, audio frame length: {len(pcm_audio_frame.data) if pcm_audio_frame.data else 0}"
            )

    # Video in
    def get_video_frames(self, uid: str) -> VideoStream | None:
        """
        Returns the video frames from the channel.

        Returns:
            AudioStream: The video stream.
        """
        return (
            None
            if self.channel_event_observer.video_streams.get(uid) is None
            else self.channel_event_observer.video_streams.get(uid)
        )

    # Video out
    async def push_yuv_video_frame(self, video_frame: ExternalVideoFrame) -> None:
        """
        Pushes an video frame to the channel.

        Parameters:
            frame: The video frame to be pushed.
        """
        frame = copy.copy(video_frame)
        ret = self.video_yuv_data_sender.send_video_frame(frame)
        logging.debug(
            f"Pushed video frame: {ret}, video frame length: {len(video_frame.buffer) if video_frame.buffer else 0}"
        )
        if ret and ret < 0:
            raise Exception(
                f"Failed to send video frame: {ret}, video frame length: {len(video_frame.buffer if video_frame.buffer else 0)}"
            )


class AgoraCallbacks(BaseModel):
    """async callback, no wait"""

    on_connected: Callable[[rtc.RTCConnection, rtc.RTCConnInfo, int], Awaitable[None]]
    on_connection_state_changed: Callable[
        [rtc.RTCConnection, rtc.RTCConnInfo, int], Awaitable[None]
    ]
    on_connection_failure: Callable[[int], Awaitable[None]]
    on_disconnected: Callable[[rtc.RTCConnection, rtc.RTCConnInfo, int], Awaitable[None]]
    on_error: Callable[[str], Awaitable[None]]
    on_participant_connected: Callable[[rtc.RTCConnection, str], Awaitable[None]]
    on_participant_disconnected: Callable[[rtc.RTCConnection, str, int], Awaitable[None]]
    on_data_received: Callable[[bytes, str], Awaitable[None]]
    on_first_participant_joined: Callable[[str], Awaitable[None]]
    on_audio_subscribe_state_changed: Callable[
        [rtc.LocalUser, str, str, int, int, int], Awaitable[None]
    ]
    on_video_subscribe_state_changed: Callable[
        [rtc.LocalUser, str, str, int, int, int], Awaitable[None]
    ]


class AgoraTransportClient:
    def __init__(
        self,
        token: str,
        params: AgoraParams,
        callbacks: AgoraCallbacks,
        loop: asyncio.AbstractEventLoop,
        service: rtc.AgoraService = None,
    ):
        self._token = token
        self._token_claims: TokenClaims = None
        if not self.verify_token():
            raise Exception(f"token {token} is not valid")

        logging.debug(f"token_claims: {self._token_claims}")
        self._app_id = self._token_claims.app_id
        self._room_name = self._token_claims.rtc.channel_name
        self._join_name = self._token_claims.rtc.uid
        self._params = params
        self._callbacks = callbacks
        self._loop = loop
        self._service = service if service else AgoraService.init(self._params)
        self._channel = RtcChannel(
            params,
            self._token_claims,
            rtc.RtcOptions(
                channel_name=self._token_claims.rtc.channel_name,
                uid=int(self._token_claims.rtc.uid) if self._token_claims.rtc.uid else 0,
                sample_rate=params.audio_in_sample_rate,
                channels=params.audio_in_channels,
                enable_pcm_dump=params.enable_pcm_dump,
            ),
            self._loop,
            self._service,
        )

        self._joined = False
        self._joining = False
        self._leaving = False

        self._other_participant_has_joined = False

        # wait to sub then notify task
        self._in_task: asyncio.Task = self._loop.create_task(self.in_handle())
        self._curr_subscribe_user: int | None = None

        # audio out
        # local participant audio stream

        # TODO: switch room anchor participant, need know room anchor
        # audio in
        # passive sub
        # self._on_participant_audio_frame_task: asyncio.Task | None = None
        # active sub
        self._in_audio_queue = asyncio.Queue[bytes]()
        self._in_audio_task: asyncio.Task = (
            self._loop.create_task(self.in_audio_handle()) if params.audio_in_enabled else None
        )

        # video in
        # passive sub the participant video frame
        self._on_participant_video_frame_task: asyncio.Task | None = None
        # active sub
        # self._in_video_queue = asyncio.Queue[bytes]()
        # self._in_video_task: asyncio.Task = self._loop.create_task(
        #    self.in_video_handle()) if params.video_in_enabled else None

        # Set up channel event handlers to call register handlers
        # NOTE: those event at last will be called for customized callback
        self._channel.on("user_joined", self.on_participant_connected)
        self._channel.on("user_left", self.on_participant_disconnected)
        self._channel.on("stream_message", self.on_data_received)
        self._channel.on("connected", self.on_connected)
        self._channel.on("disconnected", self.on_disconnected)
        self._channel.on("connection_state_changed", self.on_connection_state_changed)
        self._channel.on("connection_failure", self.on_connection_failure)
        self._channel.on("audio_subscribe_state_changed", self.on_audio_subscribe_state_changed)
        self._channel.on("video_subscribe_state_changed", self.on_video_subscribe_state_changed)

    def verify_token(self) -> bool:
        try:
            self._token_claims = TokenPaser.parse_claims(self._token)
        except Exception as e:
            logging.warning(f"verfiy {self._token} Exception: {e}")
            return False

        return True

    async def cleanup(self):
        if self._in_task and not self._in_task.cancelled():
            self._in_task.cancel()
            await self._in_task
            # logging.info("Cancelled in_task")

        if self._in_audio_task and not self._in_audio_task.cancelled():
            self._in_audio_task.cancel()
            await self._in_audio_task
            # logging.info("Cancelled in_audio_task")

        if (
            self._on_participant_video_frame_task
            and not self._on_participant_video_frame_task.cancelled()
        ):
            self._on_participant_video_frame_task.cancel()
            await self._on_participant_video_frame_task

    async def _join(self):
        self._joining = True
        participants = self.get_participant_ids()
        logging.info(
            f"{self._join_name} Connecting to {self._room_name},"
            f" current remote participants:{participants}"
        )

        try:
            await self._channel.connect()
        except Exception as e:
            self._joining = False
            self._joined = False
            raise e

        logging.info(
            f"local_participant:{self._channel.uid} joined room(channel): {self._room_name} connection_state: {self._channel.connection_state}"
        )
        self._joining = False
        self._joined = True

    async def join(self):
        # input and output procssor maybe do join room at the same time
        # need check join state, avoid duplicate join
        if self._joined or self._joining:
            logging.warning(f"{self._join_name} has connected {self._room_name}")
            return

        try:
            await self._join()

            # TODO: transcription_enabled to start

            # Set up audio source and track
            # - for pub/write out to local_participant microphone
            if self._params.audio_out_enabled:
                self._channel.set_local_user_audio_track()
                logging.info("local user audio track enabled")

            # Set up video source and track
            # for pub/write out to local_participant camera
            if self._params.camera_out_enabled:
                self._channel.set_local_user_video_track()
                logging.info("local user video track enabled")

            # Check if there are already participants in the room
            participants = self.get_participant_ids()
            if len(participants) > 0 and not self._other_participant_has_joined:
                logging.info(f"first participant {participants[0]} join")
                self._other_participant_has_joined = True
                # TODO: need check who is 主播,
                # default use the first remote participants[0],
                # or callback params use room to broadcast
                await self._callbacks.on_first_participant_joined(participants[0])

            demo_url = self._params.demo_voice_url
            if self._params.camera_in_enabled or self._params.camera_out_enabled:
                demo_url = self._params.demo_video_url
            url = f"{demo_url}?appid={quote(self._app_id)}&channel={quote(self._room_name)}&token={quote(self._token)}&uid={quote(str(self._token_claims.rtc.uid))}"
            logging.info(f"u can access url: {url}")
        except asyncio.TimeoutError:
            error_msg = f"Time out joining {self._room_name}"
            logging.error(error_msg, exc_info=True)
            await self._callbacks.on_error(error_msg)
        except Exception as e:
            error_msg = f"Error join {self._room_name}: {e}"
            logging.error(error_msg, exc_info=True)
            await self._callbacks.on_error(error_msg)

    async def _leave(self):
        # if end processor, leave room
        # need manual touch disconnect callback,some reason disconnect fail
        # await self._async_on_disconnected(reason="Leave Room.")
        self._channel.release()
        await self._channel.disconnect()

    async def leave(self):
        if not self._joined or self._leaving:
            logging.warning(f"{self._join_name} unconnect {self._room_name}, don't to leave")
            return

        self._joined = False
        self._leaving = True

        logging.info(f"{self._join_name} Disconnecting from {self._room_name}")

        # TODO: transcription_enabled to stop

        # if join and leave quick,and some message is async send message
        # need sleep to disconnect or wait completed event
        # await asyncio.sleep(1)

        try:
            await self._leave()
        except asyncio.TimeoutError:
            error_msg = f"Time out leaving {self._room_name}"
            logging.error(error_msg, exc_info=True)
            await self._callbacks.on_error(error_msg)
        except Exception as e:
            self._leaving = False
            error_msg = f"Error leave {self._room_name}: {e}"
            logging.error(error_msg, exc_info=True)
            await self._callbacks.on_error(error_msg)

    async def wait_for_remote_user(self, timeout_s: float | None = None) -> int:
        """
        if have remote users joined, return first user id;
        else wait for remote user join. (wait with timeout_s, timeout_s is None , no wait timeout)
        return user id
        """
        remote_users = self.get_participant_ids()
        if len(remote_users) > 0:
            return remote_users[0]

        future = asyncio.Future[int]()

        # do user joined callback once
        self._channel.once("user_joined", lambda conn, user_id: future.set_result(user_id))

        try:
            # Wait for the remote user with a timeout_s, timeout_s is None , no wait timeout
            remote_user = await asyncio.wait_for(future, timeout=timeout_s)
            return remote_user
        except asyncio.TimeoutError:
            future.cancel()
        except KeyboardInterrupt:
            future.cancel()
        except Exception as e:
            logging.error(f"Error waiting for remote user: {e}")
            raise

    async def send_signaling_message(self, frame: TransportMessageFrame):
        """
        TODO: send signaling message to remote_participant with rtm
        """
        if not self._joined:
            return

        try:
            pass
        except Exception as e:
            logging.error(f"Error sending data: {e}")

    async def send_message(self, frame: TransportMessageFrame):
        """
        Send a chat message in the chat channel.
        TODO: need use biz msg_id, now just use uuid
        """
        if not self._joined:
            return

        msg_id = uuid.uuid4().hex
        try:
            await self._channel.chat.send_message(
                rtc.ChatMessage(
                    message=json.dumps(frame.message),
                    # need use biz message id
                    msg_id=msg_id,
                )
            )
        except Exception as e:
            logging.error(f"Error sending data: {e}")

    def get_participant_ids(self) -> List[str]:
        return list(self._channel.remote_users.keys())

    async def get_participant_metadata(self, participant_id: str) -> dict:
        participant = self._channel.remote_users.get(participant_id)
        if participant:
            # TODO: get remote_participant_info
            return {}
        return {}

    async def subscribe(self, user_id: str):
        """subscribe audio/video
        Idempotent subscribe
        """
        try:
            if self._params.audio_in_enabled:
                await self._channel.subscribe_audio(user_id)
            if self._params.camera_in_enabled:
                await self._channel.subscribe_video(user_id)
        except Exception as e:
            logging.error(f"Error subscribing user {user_id}: {e}", exc_info=True)

    async def unsubscribe(self, user_id: str):
        """unsubscribe audio/video
        Idempotent unsubscribe, but have some error log
        """
        try:
            if self._params.audio_in_enabled:
                await self._channel.unsubscribe_audio(user_id)
        except Exception as e:
            logging.error(f"Error unsubscribe_audio user {user_id}: {e}", exc_info=True)
        try:
            if self._params.camera_in_enabled:
                await self._channel.unsubscribe_video(user_id)
        except Exception as e:
            logging.error(f"Error unsubscribe_video user {user_id}: {e}", exc_info=True)

    # Event Callback

    def on_participant_connected(self, agora_rtc_conn: rtc.RTCConnection, user_id: str):
        async def participant_connected():
            await self._callbacks.on_participant_connected(agora_rtc_conn, user_id)

        # wait
        # self._loop.run_until_complete(participant_connected())
        # no wait
        self._loop.create_task(participant_connected())

    def on_participant_disconnected(
        self, agora_rtc_conn: rtc.RTCConnection, user_id: str, reason: int
    ):
        async def participant_disconnected():
            # befor destory
            participant_ids = self.get_participant_ids()
            if len(participant_ids) > 0 and participant_ids[0] == user_id:
                await self.unsubscribe(user_id)

            self._channel.destory_when_user_left(user_id)

            # after destory
            participant_ids = self.get_participant_ids()
            if len(participant_ids) == 0:
                self._other_participant_has_joined = False
            await self._callbacks.on_participant_disconnected(agora_rtc_conn, user_id, reason)

        self._loop.create_task(participant_disconnected())

    def on_data_received(
        self,
        agora_local_user: rtc.LocalUser,
        user_id: str,
        stream_id: str,
        data: bytes,
        length: int,
    ):
        logging.info(
            f"{agora_local_user} Received stream({stream_id}) message from {user_id} with ({type(data)})length: {length}"
        )
        self._loop.create_task(self._callbacks.on_data_received(data, user_id))

    def on_connected(
        self, agora_rtc_conn: rtc.RTCConnection, conn_info: rtc.RTCConnInfo, reason: int
    ):
        self._loop.create_task(self._callbacks.on_connected(agora_rtc_conn, conn_info, reason))

    def on_connection_state_changed(
        self, agora_rtc_conn: rtc.RTCConnection, conn_info: rtc.RTCConnInfo, reason: int
    ):
        self._loop.create_task(
            self._callbacks.on_connection_state_changed(agora_rtc_conn, conn_info, reason)
        )

    def on_disconnected(
        self, agora_rtc_conn: rtc.RTCConnection, conn_info: rtc.RTCConnInfo, reason: int
    ):
        self._joined = False
        self._loop.create_task(self._callbacks.on_disconnected(agora_rtc_conn, conn_info, reason))

    def on_connection_failure(
        self, agora_rtc_conn: rtc.RTCConnection, conn_info: rtc.RTCConnInfo, reason: int
    ):
        self._joined = False
        self._loop.create_task(self._callbacks.on_connection_failure(reason))

    def on_audio_subscribe_state_changed(
        self,
        agora_local_user: rtc.LocalUser,
        channel: str,
        user_id: str,
        old_state: int,
        new_state: int,
        elapse_since_last_state: int,
    ):
        self._loop.create_task(
            self._callbacks.on_audio_subscribe_state_changed(
                agora_local_user, channel, user_id, old_state, new_state, elapse_since_last_state
            )
        )

    def on_video_subscribe_state_changed(
        self,
        agora_local_user: rtc.LocalUser,
        channel: str,
        user_id: str,
        old_state: int,
        new_state: int,
        elapse_since_last_state: int,
    ):
        self._loop.create_task(
            self._callbacks.on_video_subscribe_state_changed(
                agora_local_user, channel, user_id, old_state, new_state, elapse_since_last_state
            )
        )

    async def in_handle(self) -> None:
        """
        wait remote user to subscirbe, when no participant join;
        then notify the in audio/video handle to get audio/video stream
        """
        logging.debug("start in_handle")
        try:
            self._curr_subscribe_user = await self.wait_for_remote_user()
            logging.info(
                f"current subscribe user {self._curr_subscribe_user} other_participant_has_joined: {self._other_participant_has_joined}"
            )
            await self.subscribe(self._curr_subscribe_user)
            if not self._other_participant_has_joined:
                self._other_participant_has_joined = True
                await self._callbacks.on_first_participant_joined(self._curr_subscribe_user)
        except asyncio.CancelledError:
            logging.info("Cancelled Audio task")
        except Exception as e:
            logging.error(f"in_handle error:{e}", exc_info=True)
            return

    # Audio in

    async def in_audio_handle(self) -> None:
        """active sub the first participant audio frame"""
        logging.info("start in_audio_handle")
        try:
            # from on_playback_audio_frame_before_mixing callback
            # to get the audio frame published by the remote user
            while (
                self._curr_subscribe_user is None
                or self._channel.get_audio_frames(self._curr_subscribe_user) is None
            ):
                await asyncio.sleep(0.1)

            audio_frames = self._channel.get_audio_frames(self._curr_subscribe_user)
            logging.info(f"start get_audio_frames({self._curr_subscribe_user}): {audio_frames}")
            async for audio_frame in audio_frames:
                await self._in_audio_queue.put((audio_frame, self._curr_subscribe_user))
                # Yield control to allow other tasks to run
                await asyncio.sleep(0)
        except asyncio.CancelledError:
            logging.info("Cancelled Audio task")
            return

    async def read_next_audio_frame(self) -> AudioRawFrame | None:
        """get room sub the first participant audio frame from a queue"""
        pcm_audio_frame, participant_id = await self._in_audio_queue.get()
        audio_frame = self._convert_input_audio(pcm_audio_frame, participant_id)
        return audio_frame

    def _convert_input_audio(
        self,
        pcm_audio_frame: rtc.PcmAudioFrame,
        participant_id: str = "",
    ) -> UserAudioRawFrame | None:
        audio_data = pcm_audio_frame.data
        pcm_data = np.frombuffer(audio_data, dtype=np.int16)
        original_sample_rate = pcm_audio_frame.sample_rate
        original_num_channels = pcm_audio_frame.number_of_channels

        # Allow 8kHz and 16kHz, other sampple rate convert to 16kHz
        if original_sample_rate not in [8000, 16000]:
            sample_rate = 16000
            pcm_data = resample_audio(pcm_data, original_sample_rate, sample_rate)
        else:
            sample_rate = original_sample_rate

        if sample_rate != self._params.audio_in_sample_rate:
            self._params.audio_in_sample_rate = sample_rate
            if self._params.vad_enabled and self._params.vad_analyzer:
                self._params.vad_analyzer.set_args(
                    {
                        "sample_rate": sample_rate,
                        "num_channels": original_num_channels,
                    }
                )

        return UserAudioRawFrame(
            user_id=participant_id,
            audio=pcm_data.tobytes(),
            sample_rate=sample_rate,
            num_channels=original_num_channels,
        )

    def capture_participant_audio(
        self,
        participant_id: str,
        callback: Callable[[UserAudioRawFrame], Awaitable[None]],
        sample_rate=None,
        num_channels=None,
    ):
        """passive sub the participant audio frame"""
        # TODO: switch participant audio stream
        pass

    # Audio out

    async def write_raw_audio_frames(self, frame_data: bytes):
        if not self._joined:
            return

        try:
            pcm_audio_frame: rtc.PcmAudioFrame = self._convert_output_audio(frame_data)
            await self._channel.push_pcm_audio_frame(pcm_audio_frame)

            # TODONE: need fix push_audio_frame,
            # the sample_rate is not correct, need use output_sample_rate
            # await self._channel.push_audio_frame(frame_data)
        except Exception as e:
            logging.error(f"Error publishing audio: {e}", exc_info=True)

    def _convert_output_audio(self, audio_data: bytes) -> rtc.PcmAudioFrame:
        bytes_per_sample = SAMPLE_WIDTH  # Assuming 16-bit audio sample_width
        total_samples = len(audio_data) // bytes_per_sample
        samples_per_channel = total_samples // self._params.audio_out_channels

        frame = rtc.PcmAudioFrame()
        frame.data = bytearray(audio_data)
        frame.timestamp = 0
        frame.bytes_per_sample = bytes_per_sample
        frame.samples_per_channel = samples_per_channel
        frame.sample_rate = self._params.audio_out_sample_rate
        frame.number_of_channels = self._params.audio_out_channels

        return frame

    # Camera in

    async def in_video_handle(self) -> None:
        """active sub the first participant video frame"""
        logging.info("start in_video_handle")
        try:
            # from on_frame callback
            # to get the video frame published by the remote user
            while (
                self._curr_subscribe_user is None
                or self._channel.get_video_frames(self._curr_subscribe_user) is None
            ):
                await asyncio.sleep(0.1)

            video_frames = self._channel.get_video_frames(self._curr_subscribe_user)
            logging.info(f"start get_video_frames({self._curr_subscribe_user}): {video_frames}")
            async for video_frame in video_frames:
                await self._in_video_queue.put((video_frame, self._curr_subscribe_user))
                # Yield control to allow other tasks to run
                await asyncio.sleep(0)
        except asyncio.CancelledError:
            logging.info("Cancelled video task")
            return

    def _convert_input_video_image(
        self, participant_id: str, video_frame: VideoFrame, target_color_mode: str = "RGB"
    ) -> UserImageRawFrame | None:
        """Convert input video frame to image
        target_color_mode from PIL.Image.Image convert method mode param

        cpu/gpu binding optimization:)
        - now just use opencv to optimize the performance, when use cpu (c/c++)
        - SIMD Vectorization
        - GPU Acceleration
        - use Numba jit to run fast
        """
        match video_frame.type:
            case const.AGORA_VIDEO_PIXEL_I420:  # default use yuv420
                # too slow use python lib 64 ms with cpu, need use GPU optimized
                # image = convert_I420_to_RGB(video_frame)
                image = convert_I420_to_RGB_with_cv(video_frame)  # cv use c/c++ lib 2 ms with cpu
            # case const.AGORA_VIDEO_PIXEL_RGBA:
            # need to check RGBA from video
            #    image = convert_RGBA_to_RGB(video_frame)
            case const.AGORA_VIDEO_PIXEL_NV21:
                # image = convert_NV21_to_RGB(video_frame)
                image = convert_NV21_to_RGB_optimized(video_frame)
            case const.AGORA_VIDEO_PIXEL_I422:
                # image = convert_I422_to_RGB(video_frame)
                image = convert_I422_to_RGB_with_cv(video_frame)
            case const.AGORA_VIDEO_PIXEL_NV12:
                # image = convert_NV12_to_RGB(video_frame)
                image = convert_NV12_to_RGB_optimized(video_frame)
            case _:
                logging.warning(f"buffer type:{video_frame.type} un support convert")
                return None

        if target_color_mode != "RGB":
            image = image.convert(target_color_mode)

        return UserImageRawFrame(
            user_id=participant_id,
            image=image.tobytes(),
            size=(video_frame.width, video_frame.height),
            mode=target_color_mode,
            format="JPEG",
        )

    def capture_participant_video(
        self,
        participant_id: str,
        callback: Callable[[UserImageRawFrame], Awaitable[None]],
        framerate: int = 30,
        video_source: str = "camera",
        color_format: str = "RGB",
    ):
        """passive sub the participant video frame"""
        video_stream = self._channel.get_video_frames(participant_id)
        if not video_stream:
            logging.error(f"participant_id {participant_id} no video stream")
            return

        self._on_participant_video_frame_task = asyncio.create_task(
            self._async_on_participant_video_frame(
                participant_id, callback, video_stream, color_format
            )
        )

    async def _async_on_participant_video_frame(
        self,
        participant_id: str,
        callback: Callable[[UserImageRawFrame], Awaitable[None]],
        video_stream: VideoStream,
        color_format: str = "RGB",
    ):
        logging.info(
            f"Started capture participant_id:{participant_id} from video stream {video_stream}"
        )
        async for video_frame in video_stream:
            try:
                start = time.time() * 1000
                # covert video yuv frame to image (yuv->rgb)
                image_frame = self._convert_input_video_image(
                    participant_id,
                    video_frame,
                    target_color_mode=color_format,
                )
                logging.debug(f"convert_input_video_image time: {time.time()*1000 - start} ms")

                start = time.time() * 1000
                if asyncio.iscoroutinefunction(callback):
                    await callback(image_frame)
                else:
                    callback(image_frame)
                logging.debug(f"callback time: {time.time()*1000 - start} ms")
            except asyncio.CancelledError:
                logging.info("task cancelled")
                break
            except Exception as e:
                logging.error(f"task Error: {e}", exc_info=True)

    # Camera out

    async def write_frame_to_camera(self, frame: ImageRawFrame):
        """
        publish image frame to camera
        !NOTE: (now just support RGBA color_format/mode)
        """
        try:
            image = PIL.Image.frombytes(frame.mode, frame.size, frame.image).convert("RGBA")
            # image.save(os.path.join(VIDEOS_DIR, "temp.png"))

            frame_buffer = bytearray(image.tobytes())
            logging.debug(
                f"width:{image.width}, height:{image.height}"
                f" rgba_len:{image.width * image.height * 4}, len:{len(frame_buffer)}"
            )

            video_frame = ExternalVideoFrame()
            video_frame.type = const.AGORA_VIDEO_BUFFER_RAW_DATA
            video_frame.format = const.AGORA_VIDEO_PIXEL_RGBA
            video_frame.buffer = frame_buffer
            video_frame.stride = image.width
            video_frame.height = image.height
            video_frame.timestamp = 0
            video_frame.metadata = bytearray(b"from achatbot ImageRawFrame")

            await self._channel.push_yuv_video_frame(video_frame)
        except Exception as e:
            logging.error(f"task Error: {e}", exc_info=True)



import os
import logging
import time

from src.common.interface import IBot
from . import register_ai_room_bots


@register_ai_room_bots.register
class DummyBot(IBot):
    def __init__(self, **args) -> None:
        self.args = args

    def run(self):
        logging.info(f"Starting bot,env: {os.environ}")
        logging.info(f"dummy bot run with args: {self.args}")
        time.sleep(10)
        logging.info("dummy bot over")

    async def arun(self):
        pass

    def bot_config(self) -> dict:
        return {}



import torch
import torch.nn.functional as F
import torch.nn as nn
from torch.nn import Conv1d, AvgPool1d, Conv2d
from torch.nn.utils import weight_norm, spectral_norm

from .utils import get_padding

LRELU_SLOPE = 0.1

def stft(x, fft_size, hop_size, win_length, window):
    """Perform STFT and convert to magnitude spectrogram.
    Args:
        x (Tensor): Input signal tensor (B, T).
        fft_size (int): FFT size.
        hop_size (int): Hop size.
        win_length (int): Window length.
        window (str): Window function type.
    Returns:
        Tensor: Magnitude spectrogram (B, #frames, fft_size // 2 + 1).
    """
    x_stft = torch.stft(x, fft_size, hop_size, win_length, window,
            return_complex=True)
    real = x_stft[..., 0]
    imag = x_stft[..., 1]

    return torch.abs(x_stft).transpose(2, 1)

class SpecDiscriminator(nn.Module):
    """docstring for Discriminator."""

    def __init__(self, fft_size=1024, shift_size=120, win_length=600, window="hann_window", use_spectral_norm=False):
        super(SpecDiscriminator, self).__init__()
        norm_f = weight_norm if use_spectral_norm == False else spectral_norm
        self.fft_size = fft_size
        self.shift_size = shift_size
        self.win_length = win_length
        self.window = getattr(torch, window)(win_length)
        self.discriminators = nn.ModuleList([
            norm_f(nn.Conv2d(1, 32, kernel_size=(3, 9), padding=(1, 4))),
            norm_f(nn.Conv2d(32, 32, kernel_size=(3, 9), stride=(1,2), padding=(1, 4))),
            norm_f(nn.Conv2d(32, 32, kernel_size=(3, 9), stride=(1,2), padding=(1, 4))),
            norm_f(nn.Conv2d(32, 32, kernel_size=(3, 9), stride=(1,2), padding=(1, 4))),
            norm_f(nn.Conv2d(32, 32, kernel_size=(3, 3), stride=(1,1), padding=(1, 1))),
        ])

        self.out = norm_f(nn.Conv2d(32, 1, 3, 1, 1))

    def forward(self, y):

        fmap = []
        y = y.squeeze(1)
        y = stft(y, self.fft_size, self.shift_size, self.win_length, self.window.to(y.get_device()))
        y = y.unsqueeze(1)
        for i, d in enumerate(self.discriminators):
            y = d(y)
            y = F.leaky_relu(y, LRELU_SLOPE)
            fmap.append(y)

        y = self.out(y)
        fmap.append(y)

        return torch.flatten(y, 1, -1), fmap

class MultiResSpecDiscriminator(torch.nn.Module):

    def __init__(self,
                 fft_sizes=[1024, 2048, 512],
                 hop_sizes=[120, 240, 50],
                 win_lengths=[600, 1200, 240],
                 window="hann_window"):

        super(MultiResSpecDiscriminator, self).__init__()
        self.discriminators = nn.ModuleList([
            SpecDiscriminator(fft_sizes[0], hop_sizes[0], win_lengths[0], window),
            SpecDiscriminator(fft_sizes[1], hop_sizes[1], win_lengths[1], window),
            SpecDiscriminator(fft_sizes[2], hop_sizes[2], win_lengths[2], window)
            ])

    def forward(self, y, y_hat):
        y_d_rs = []
        y_d_gs = []
        fmap_rs = []
        fmap_gs = []
        for i, d in enumerate(self.discriminators):
            y_d_r, fmap_r = d(y)
            y_d_g, fmap_g = d(y_hat)
            y_d_rs.append(y_d_r)
            fmap_rs.append(fmap_r)
            y_d_gs.append(y_d_g)
            fmap_gs.append(fmap_g)

        return y_d_rs, y_d_gs, fmap_rs, fmap_gs


class DiscriminatorP(torch.nn.Module):
    def __init__(self, period, kernel_size=5, stride=3, use_spectral_norm=False):
        super(DiscriminatorP, self).__init__()
        self.period = period
        norm_f = weight_norm if use_spectral_norm == False else spectral_norm
        self.convs = nn.ModuleList([
            norm_f(Conv2d(1, 32, (kernel_size, 1), (stride, 1), padding=(get_padding(5, 1), 0))),
            norm_f(Conv2d(32, 128, (kernel_size, 1), (stride, 1), padding=(get_padding(5, 1), 0))),
            norm_f(Conv2d(128, 512, (kernel_size, 1), (stride, 1), padding=(get_padding(5, 1), 0))),
            norm_f(Conv2d(512, 1024, (kernel_size, 1), (stride, 1), padding=(get_padding(5, 1), 0))),
            norm_f(Conv2d(1024, 1024, (kernel_size, 1), 1, padding=(2, 0))),
        ])
        self.conv_post = norm_f(Conv2d(1024, 1, (3, 1), 1, padding=(1, 0)))

    def forward(self, x):
        fmap = []

        # 1d to 2d
        b, c, t = x.shape
        if t % self.period != 0: # pad first
            n_pad = self.period - (t % self.period)
            x = F.pad(x, (0, n_pad), "reflect")
            t = t + n_pad
        x = x.view(b, c, t // self.period, self.period)

        for l in self.convs:
            x = l(x)
            x = F.leaky_relu(x, LRELU_SLOPE)
            fmap.append(x)
        x = self.conv_post(x)
        fmap.append(x)
        x = torch.flatten(x, 1, -1)

        return x, fmap


class MultiPeriodDiscriminator(torch.nn.Module):
    def __init__(self):
        super(MultiPeriodDiscriminator, self).__init__()
        self.discriminators = nn.ModuleList([
            DiscriminatorP(2),
            DiscriminatorP(3),
            DiscriminatorP(5),
            DiscriminatorP(7),
            DiscriminatorP(11),
        ])

    def forward(self, y, y_hat):
        y_d_rs = []
        y_d_gs = []
        fmap_rs = []
        fmap_gs = []
        for i, d in enumerate(self.discriminators):
            y_d_r, fmap_r = d(y)
            y_d_g, fmap_g = d(y_hat)
            y_d_rs.append(y_d_r)
            fmap_rs.append(fmap_r)
            y_d_gs.append(y_d_g)
            fmap_gs.append(fmap_g)

        return y_d_rs, y_d_gs, fmap_rs, fmap_gs
    
class WavLMDiscriminator(nn.Module):
    """docstring for Discriminator."""

    def __init__(self, slm_hidden=768, 
                 slm_layers=13, 
                 initial_channel=64, 
                 use_spectral_norm=False):
        super(WavLMDiscriminator, self).__init__()
        norm_f = weight_norm if use_spectral_norm == False else spectral_norm
        self.pre = norm_f(Conv1d(slm_hidden * slm_layers, initial_channel, 1, 1, padding=0))
        
        self.convs = nn.ModuleList([
            norm_f(nn.Conv1d(initial_channel, initial_channel * 2, kernel_size=5, padding=2)),
            norm_f(nn.Conv1d(initial_channel * 2, initial_channel * 4, kernel_size=5, padding=2)),
            norm_f(nn.Conv1d(initial_channel * 4, initial_channel * 4, 5, 1, padding=2)),
        ])

        self.conv_post = norm_f(Conv1d(initial_channel * 4, 1, 3, 1, padding=1))
        
    def forward(self, x):
        x = self.pre(x)
        
        fmap = []
        for l in self.convs:
            x = l(x)
            x = F.leaky_relu(x, LRELU_SLOPE)
            fmap.append(x)
        x = self.conv_post(x)
        x = torch.flatten(x, 1, -1)

        return x


from time import perf_counter
import time
from typing import Optional
import modal
import os

app = modal.App("qwen2_5_omni")
omni_img = (
    # https://catalog.ngc.nvidia.com/orgs/nvidia/containers/cuda/tags
    modal.Image.from_registry(
        "nvidia/cuda:12.6.1-cudnn-devel-ubuntu22.04",
        add_python="3.10",
    )
    .apt_install("git", "git-lfs", "ffmpeg", "clang", "cmake")
    .pip_install("wheel", "openai", "qwen-omni-utils[decord]")
    .pip_install(
        "accelerate",
        "torch==2.6.0",
        "torchaudio==2.6.0",
        "torchvision==0.21.0",
        "soundfile==0.13.0",
        "librosa==0.11.0",
    )
    .run_commands(
        "pip install git+https://github.com/huggingface/transformers@v4.51.3-Qwen2.5-Omni-preview"
    )
    .pip_install("flash-attn", extra_options="--no-build-isolation")
    .env(
        {
            "PYTORCH_CUDA_ALLOC_CONF": "expandable_segments:True",
        }
    )
)

achatbot_version = os.getenv("ACHATBOT_VERSION", "")
if achatbot_version:
    omni_img = (
        omni_img.pip_install(
            f"achatbot[llm_transformers_manual_vision_voice_qwen]=={achatbot_version}",
            extra_index_url=os.getenv("EXTRA_INDEX_URL", "https://pypi.org/simple/"),
        )
        # .pip_install("flash-attn==2.5.8", extra_options="--no-build-isolation")
        .env(
            {
                "ACHATBOT_PKG": "1",
                "LOG_LEVEL": os.getenv("LOG_LEVEL", "info"),
            }
        )
    )

HF_MODEL_DIR = "/root/.achatbot/models"
hf_model_vol = modal.Volume.from_name("models", create_if_missing=True)
ASSETS_DIR = "/root/.achatbot/assets"
assets_dir = modal.Volume.from_name("assets", create_if_missing=True)

# NOTE: if want to generate speech, need use this system prompt to generate speech
SPEECH_SYS_PROMPT = "You are Qwen, a virtual human developed by the Qwen Team, Alibaba Group, capable of perceiving auditory and visual inputs, as well as generating text and speech."
# Voice settings
SPEAKER_LIST = ["Chelsie", "Ethan"]
DEFAULT_SPEAKER = "Ethan"

with omni_img.imports():
    import subprocess
    from threading import Thread
    from queue import Queue
    import numpy as np
    from transformers.generation.streamers import BaseStreamer

    import torch
    from transformers import (
        Qwen2_5OmniForConditionalGeneration,
        Qwen2_5OmniProcessor,
        TextIteratorStreamer,
        AutoConfig,
        AutoProcessor,
    )
    from qwen_omni_utils import process_mm_info

    def print_model_params(model: torch.nn.Module, extra_info=""):
        # print the number of parameters in the model
        model_million_params = sum(p.numel() for p in model.parameters()) / 1e6
        # print(model)
        print(f"{extra_info} {model_million_params} M parameters")

    class Qwen2_5OmniForConditionalGenerationNew(Qwen2_5OmniForConditionalGeneration):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            print_model_params(self.thinker, "qwen2.5omni_thinker")
            print_model_params(self.talker, "qwen2.5omni_talker")
            print_model_params(self.token2wav, "qwen2.5omni_token2wav")

        @torch.no_grad()
        # TODO: raushan, defaults should be saved in generation config
        def generate(
            self,
            input_ids: Optional[torch.Tensor] = None,
            speaker: str = "Chelsie",
            use_audio_in_video: bool = False,
            return_audio: Optional[bool] = None,
            thinker_max_new_tokens: int = 1024,
            talker_max_new_tokens: int = 4096,
            talker_do_sample: bool = True,
            talker_top_k: int = 40,
            talker_top_p: float = 0.8,
            talker_temperature: float = 0.9,
            talker_eos_token_id: list[int] = [8292, 8294],
            talker_repetition_penalty: float = 1.05,
            **kwargs,
        ):
            r"""
            Generate text response and audio from input.

            Args:
                input_ids (`Optional[torch.Tensor]`, *optional*):
                    Input ids, should obtain from processor.
                speaker (`str` , defaults to "Chelsie"):
                    Which speaker should be used in audio response.
                use_audio_in_video (`bool`, defaults to False):
                    Whether or not use audio track in video, should same as the parameter in `process_audio_info`.
                return_audio (`Optional[bool]`, *optional*):
                    Whether or not return response in audio format. When `return_audio=None`, this parameter is same as `config.enable_audio_output`.
                kwargs (*optional*):
                    - Without a prefix, they will be entered as `**kwargs` for the `generate` method of each sub-model.
                    - With a *thinker_*, *talker_*, *token2wav_* prefix, they will be input for the `generate` method of the
                    thinker, talker and token2wav respectively. It has the priority over the keywords without a prefix.
            Returns:
                When `return_audio=False`:
                    - **Text** (`torch.Tensor`): Generated text token sequence.
                When `return_audio=True`:
                    - **Text** (`torch.Tensor`): Generated text token sequence.
                    - **Audio waveform** (`torch.Tensor`): Generated audio waveform.
            """
            if speaker not in self.speaker_map:
                raise ValueError(
                    f"{speaker} is not availible, availible speakers: {self.speaker_map.keys()}"
                )
            if return_audio and not self.has_talker:
                raise ValueError(
                    "Cannot use talker when talker module not initalized. Use `enable_talker` method or set enable_talker in config to enable talker."
                )
            if return_audio is None:
                return_audio = self.has_talker
            if input_ids.shape[0] != 1 and return_audio:
                raise NotImplementedError(
                    "Qwen2.5-Omni currently does not support batched inference with audio output"
                )

            shared_kwargs = {"use_audio_in_video": use_audio_in_video}
            thinker_kwargs = {
                "max_new_tokens": thinker_max_new_tokens,
            }
            talker_kwargs = {
                "max_new_tokens": talker_max_new_tokens,
                "do_sample": talker_do_sample,
                "top_k": talker_top_k,
                "top_p": talker_top_p,
                "temperature": talker_temperature,
                "eos_token_id": talker_eos_token_id,
                "repetition_penalty": talker_repetition_penalty,
            }
            token2wav_kwargs = {}

            for key, value in kwargs.items():
                if key.startswith("thinker_"):
                    thinker_kwargs[key[len("thinker_") :]] = value
                elif key.startswith("talker_"):
                    talker_kwargs[key[len("talker_") :]] = value
                elif key.startswith("token2wav_"):
                    token2wav_kwargs[key[len("token2wav_") :]] = value
                # Process special input values
                elif key == "feature_attention_mask":
                    thinker_kwargs[key] = value
                    talker_kwargs["audio_feature_lengths"] = torch.sum(value, dim=1)
                elif key == "input_features" or key == "attention_mask":
                    thinker_kwargs[key] = value
                # Put other key to shared kwargs
                else:
                    shared_kwargs[key] = value

            # Merge kwargs
            for key, value in shared_kwargs.items():
                if key not in thinker_kwargs:
                    thinker_kwargs[key] = value
                if key not in talker_kwargs:
                    talker_kwargs[key] = value
                if key not in token2wav_kwargs:
                    token2wav_kwargs[key] = value
            speaker_params = self.speaker_map[speaker]

            # 1. Generate from thinker module
            generate_audio = return_audio and self.has_talker
            if generate_audio:
                thinker_kwargs["output_hidden_states"] = True
                thinker_kwargs["return_dict_in_generate"] = True

            thinker_result = self.thinker.generate(input_ids=input_ids, **thinker_kwargs)

            if not generate_audio:
                return thinker_result

            # 2. Generate speech tokens from talker module
            thinker_generate_ids = thinker_result.sequences[:, input_ids.size(1) :].to(
                self.talker.device
            )
            thinker_token_embeds = [
                token_hidden_states[0].to(self.talker.device)
                for token_hidden_states in thinker_result.hidden_states
            ]
            thinker_hidden_states = [
                token_hidden_states[-1].to(self.talker.device)
                for token_hidden_states in thinker_result.hidden_states
            ]

            talker_text_bos_token = speaker_params["bos_token"]
            talker_input_text_ids = torch.cat(
                [
                    input_ids.to(self.talker.device),
                    torch.tensor(
                        [[talker_text_bos_token]], dtype=torch.long, device=self.talker.device
                    ),
                    thinker_generate_ids[:, :1],
                ],
                dim=-1,
            )

            talker_input_ids = torch.cat(
                [
                    torch.full_like(
                        input_ids,
                        fill_value=self.talker.codec_mask_token,
                        device=self.talker.device,
                    ),
                    torch.tensor(
                        [[self.talker.codec_pad_token]], dtype=torch.long, device=self.talker.device
                    ),
                    torch.tensor(
                        [[self.talker.codec_bos_token]], dtype=torch.long, device=self.talker.device
                    ),
                ],
                dim=1,
            )

            thinker_embed_tokens = self.thinker.get_input_embeddings()
            thinker_reply_part = torch.cat(thinker_hidden_states[1:], dim=1) + torch.cat(
                thinker_token_embeds[1:], dim=1
            )
            talker_inputs_embeds = thinker_hidden_states[0] + thinker_token_embeds[0]
            talker_text_bos_token = torch.tensor(
                [[talker_text_bos_token]], dtype=torch.long, device=self.thinker.device
            )
            talker_text_bos_embed = thinker_embed_tokens(talker_text_bos_token).to(
                self.talker.device
            )
            talker_inputs_embeds = torch.cat(
                [
                    talker_inputs_embeds,
                    talker_text_bos_embed,
                    thinker_reply_part[:, :1, :],
                ],
                dim=1,
            )

            eos_embedding = thinker_embed_tokens(
                torch.tensor(
                    [[self.talker.text_eos_token]], dtype=torch.long, device=self.thinker.device
                )
            ).to(self.talker.device)

            pad_embedding = thinker_embed_tokens(
                torch.tensor(
                    [[self.talker.text_pad_token]], dtype=torch.long, device=self.thinker.device
                )
            ).to(self.talker.device)

            thinker_reply_part = torch.cat(
                [
                    thinker_reply_part[:, 1:, :],
                    eos_embedding,
                    pad_embedding,
                ],
                dim=1,
            )

            talker_attention_mask = None
            if "attention_mask" in kwargs:
                talker_attention_mask = torch.cat(
                    [kwargs["attention_mask"], kwargs["attention_mask"].new_ones((1, 2))], dim=1
                ).to(self.talker.device)

            # stream
            skip_prompt = kwargs.get("skip_prompt", True)
            streamer = TokenStreamer(skip_prompt=skip_prompt)
            talker_kwargs = dict(
                input_ids=talker_input_ids,
                streamer=streamer,
                input_text_ids=talker_input_text_ids,
                thinker_reply_part=thinker_reply_part,
                inputs_embeds=talker_inputs_embeds,
                attention_mask=talker_attention_mask,
                suppress_tokens=[self.talker.codec_bos_token],
                **{
                    k: (v.to(self.talker.device) if torch.is_tensor(v) else v)
                    for k, v in talker_kwargs.items()
                },
            )
            # print(talker_kwargs.keys())
            thread = Thread(target=self.talker.generate, kwargs=talker_kwargs)
            thread.start()
            talker_generate_codes = []
            times = []
            start_time = perf_counter()
            for token_id in streamer:
                # print(token_id)
                times.append(perf_counter() - start_time)
                start_time = perf_counter()
                talker_generate_codes.append(token_id)
            print(
                f"generate first token cost time: {times[0]} s, {len(times)} tokens cost time: {sum(times)} s"
            )
            offset = 0
            if skip_prompt is False:
                offset = talker_input_ids.shape[1]
            # print(
            #    talker_input_ids.shape[1],
            #    # talker_generate_codes,
            #    talker_generate_codes[:offset],
            #    talker_generate_codes[offset:-1],
            # )
            talker_generate_codes = torch.tensor(
                [talker_generate_codes[offset:-1]],
                dtype=torch.long,
                device=self.talker.device,
            )

            # no stream
            # talker_result = self.talker.generate(
            #     input_ids=talker_input_ids,
            #     input_text_ids=talker_input_text_ids,
            #     thinker_reply_part=thinker_reply_part,
            #     inputs_embeds=talker_inputs_embeds,
            #     attention_mask=talker_attention_mask,
            #     suppress_tokens=[self.talker.codec_bos_token],
            #     **{
            #         k: (v.to(self.talker.device) if torch.is_tensor(v) else v)
            #         for k, v in talker_kwargs.items()
            #     },
            # )
            # print(talker_result.shape, talker_result)
            # talker_generate_codes = talker_result[:, talker_input_ids.shape[1] : -1]

            # print(f"talker_generate_codes:{talker_generate_codes.shape} {talker_generate_codes}")

            # 3. Generate wavs from code
            if self.token2wav.dtype != torch.float:
                self.token2wav.float()

            # print(self.token2wav.device, speaker_params, token2wav_kwargs)

            wav = self.token2wav(
                talker_generate_codes.to(self.token2wav.device),
                conditioning=speaker_params["cond"].to(self.token2wav.device).float(),
                reference_mel=speaker_params["ref_mel"].to(self.token2wav.device).float(),
                **token2wav_kwargs,
            )

            return thinker_result.sequences, wav.float()

    subprocess.run("nvidia-smi --version", shell=True)
    subprocess.run("nvcc --version", shell=True)
    gpu_prop = torch.cuda.get_device_properties("cuda")
    print(gpu_prop)

    if not os.getenv("ACHATBOT_PKG"):
        model_path = os.path.join(HF_MODEL_DIR, "Qwen/Qwen2.5-Omni-7B")
        config = AutoConfig.from_pretrained(model_path)
        model = Qwen2_5OmniForConditionalGenerationNew.from_pretrained(
            model_path,
            torch_dtype=torch.bfloat16,
            device_map={"": 0},
            attn_implementation="flash_attention_2",
            config=config,
        ).eval()

        # NOTE: when disable talker, generate must set return_audio=False
        # model.disable_talker()

        print_model_params(model, "Qwen2.5Omni")

        # processor = Qwen2_5OmniProcessor.from_pretrained(model_path)
        processor = AutoProcessor.from_pretrained(
            model_path,
            min_pixels=256 * 28 * 28,
            max_pixels=1280 * 28 * 28,
            trust_remote_code=True,
        )

    # subprocess.run("nvidia-smi", shell=True)

    def inference(
        messages,
        return_audio=False,
        use_audio_in_video=False,
        thinker_do_sample=False,
        speaker=DEFAULT_SPEAKER,
    ):
        text = processor.apply_chat_template(messages, tokenize=False, add_generation_prompt=True)
        # image_inputs, video_inputs = process_vision_info([messages])
        audios, images, videos = process_mm_info(messages, use_audio_in_video=use_audio_in_video)
        inputs = processor(
            text=text,
            audio=audios,
            images=images,
            videos=videos,
            return_tensors="pt",
            padding=True,
            use_audio_in_video=use_audio_in_video,
        )
        inputs = inputs.to(model.device).to(model.dtype)
        for k, v in inputs.items():
            print(k, v.shape)

        output = model.generate(
            **inputs,
            use_audio_in_video=use_audio_in_video,
            return_audio=return_audio,
            speaker=speaker,
            thinker_do_sample=thinker_do_sample,
        )
        print("\n====generate use memory=====\n")
        subprocess.run(
            """nvidia-smi --query-gpu=index,memory.used,memory.total --format=csv,noheader,nounits | awk -F',' '{print "GPU "$1": "$2"/"$3" MiB\\n"}'""",
            shell=True,
        )
        print("\n=========\n")
        # print(output)
        text_token_ids = output
        audio = None
        if return_audio and len(output) > 1:
            text_token_ids = output[0].detach()
            audio = output[1].unsqueeze(0).detach()

        text = processor.batch_decode(
            text_token_ids, skip_special_tokens=True, clean_up_tokenization_spaces=False
        )

        torch.cuda.empty_cache()

        return text, audio

    def thinker_inference_stream(
        messages,
        use_audio_in_video=False,
    ):
        print(messages)
        text = processor.apply_chat_template(messages, tokenize=False, add_generation_prompt=True)
        # image_inputs, video_inputs = process_vision_info([messages])
        audios, images, videos = process_mm_info(messages, use_audio_in_video=use_audio_in_video)
        print(text)
        {print(f"audios[{i}]: {item.shape}") for i, item in enumerate(audios)} if audios else print(
            audios
        )
        {print(f"images[{i}]: {item}") for i, item in enumerate(images)} if images else print(
            images
        )
        {print(f"videos[{i}]: {item.shape}") for i, item in enumerate(videos)} if videos else print(
            videos
        )

        inputs = processor(
            text=text,
            audio=audios,
            images=images,
            videos=videos,
            return_tensors="pt",
            padding=True,
            use_audio_in_video=use_audio_in_video,
        )
        inputs = inputs.to(model.device).to(model.dtype)
        for k, v in inputs.items():
            print(k, v.shape)

        streamer = TextIteratorStreamer(processor, skip_prompt=True, skip_special_tokens=True)

        generation_kwargs = dict(
            **inputs,
            streamer=streamer,
            use_audio_in_video=use_audio_in_video,
            return_audio=False,
            thinker_do_sample=True,
            # do_sample=True,
            top_k=20,
            top_p=0.8,
            temperature=0.1,
            repetition_penalty=1.0,
            min_new_tokens=0,
            max_new_tokens=1024,
        )
        thread = Thread(target=model.generate, kwargs=generation_kwargs)
        thread.start()

        generated_text = ""
        times = []
        start_time = perf_counter()
        for new_text in streamer:
            times.append(perf_counter() - start_time)
            start_time = perf_counter()
            generated_text += new_text
            yield new_text
        print(
            f"generate [{generated_text}] first token cost time: {times[0]} s, {len(times)} tokens cost time: {sum(times)} s"
        )
        torch.cuda.empty_cache()

    def thinker_inference_chunk_stream(
        messages,
        use_audio_in_video=False,
        max_new_tokens=2048,
        max_tokens_per_step=3,  # Controls how many tokens to generate *per step*
        eos_token_ids=[151644, 151645],  # Define EOS tokens
        output_hidden_states=False,
    ):
        print(messages)
        text = processor.apply_chat_template(messages, tokenize=False, add_generation_prompt=True)
        # image_inputs, video_inputs = process_vision_info([messages])
        audios, images, videos = process_mm_info(messages, use_audio_in_video=use_audio_in_video)
        print(text)
        {print(f"audios[{i}]: {item.shape}") for i, item in enumerate(audios)} if audios else print(
            audios
        )
        {print(f"images[{i}]: {item}") for i, item in enumerate(images)} if images else print(
            images
        )
        {print(f"videos[{i}]: {item.shape}") for i, item in enumerate(videos)} if videos else print(
            videos
        )

        inputs = processor(
            text=text,
            audio=audios,
            images=images,
            videos=videos,
            return_tensors="pt",
            padding=True,
            use_audio_in_video=use_audio_in_video,
        )
        inputs = inputs.to(model.device).to(model.dtype)
        for k, v in inputs.items():
            print(k, v.shape)
        """
        e.g.:
            input_ids torch.Size([1, 20174])
            attention_mask torch.Size([1, 20174])

            pixel_values_videos torch.Size([77760, 1176])
            image_grid_thw torch.Size([1, 3]) # just image only

            video_grid_thw torch.Size([1, 3])
            video_second_per_grid torch.Size([1])

            feature_attention_mask torch.Size([1, 30000])
            input_features torch.Size([1, 128, 30000])
        """
        return thinker_generate_chunk(
            inputs,
            max_new_tokens=max_new_tokens,
            max_tokens_per_step=max_tokens_per_step,
            use_audio_in_video=use_audio_in_video,
            eos_token_ids=eos_token_ids,
            output_hidden_states=output_hidden_states,
        )

    @torch.no_grad()
    def thinker_generate_chunk(
        inputs: dict,
        max_new_tokens=2048,
        max_tokens_per_step=10,  # Controls how many tokens to generate *per step*
        use_audio_in_video=False,
        eos_token_ids=[151644, 151645],  # Define EOS tokens
        output_hidden_states=False,
        stop_strings_per_step=[".", "。"],
    ):
        input_ids = inputs.pop("input_ids")
        attention_mask = inputs.pop("attention_mask", None)

        if max_tokens_per_step > max_new_tokens:
            max_tokens_per_step = max_new_tokens

        # Keep track of the full generated sequence full_generated_ids = input_ids.clone()
        # Ensure full_attention_mask is correctly initialized and expanded
        full_attention_mask = (
            attention_mask.clone()
            if attention_mask is not None
            else torch.ones_like(input_ids, device=input_ids.device)
        )
        full_generated_ids = input_ids.clone()

        # KV cache
        # past_key_values = None

        # Inputs for the current step
        current_input_ids = full_generated_ids
        # The attention mask passed to generate should cover the sequence length for the current step
        current_attention_mask = full_attention_mask

        total_new_tokens_generated = 0
        generated_text = ""
        hidden_states = None
        hidden_states_len = 0

        times = []
        while total_new_tokens_generated < max_new_tokens:
            # Prepare inputs for generate call
            # print(current_input_ids, current_attention_mask.shape)
            # https://huggingface.co/docs/transformers/v4.51.3/en/main_classes/text_generation#transformers.GenerationMixin.generate
            model_inputs = {
                "input_ids": current_input_ids,
                "attention_mask": current_attention_mask,
                # "past_key_values": past_key_values,
                "use_cache": True,
                "use_audio_in_video": use_audio_in_video,
                "do_sample": True,
                "top_k": 10,
                "top_p": 0.9,
                "temperature": 0.95,
                "repetition_penalty": 1.1,
                "min_new_tokens": 1,  # Ensure at least one token is generated if possible
                "max_new_tokens": max_tokens_per_step,  # Generate in smaller steps
                # output_hidden_states/scores can consume memory,
                # enable if needed downstream(talker)
                "output_hidden_states": output_hidden_states,
                "return_dict_in_generate": True,
                # "output_scores": True,
                "eos_token_id": eos_token_ids,
                "pad_token_id": processor.tokenizer.pad_token_id,
            }
            model_inputs = {**inputs, **model_inputs}
            for k, v in model_inputs.items():
                if isinstance(v, torch.Tensor):
                    print(k, v.shape)
                else:
                    print(k, v)
            if len(stop_strings_per_step) > 0:
                model_inputs["stop_strings"] = stop_strings_per_step
                model_inputs["tokenizer"] = processor.tokenizer

            start_time = perf_counter()
            outputs = model.thinker.generate(**model_inputs)
            times.append(perf_counter() - start_time)

            # Extract newly generated token IDs *for this step*
            # `outputs.sequences` contains the input_ids for this step + new tokens generated in this step
            step_new_ids = outputs.sequences[:, current_input_ids.shape[1] :]
            num_step_new_tokens = step_new_ids.shape[1]

            if num_step_new_tokens == 0:  # Handle case where generate stops early
                print("Warning: generate produced 0 new tokens in this step.")
                break

            if output_hidden_states is True:
                hidden_states = outputs.hidden_states
                print(hidden_states[0][0].shape)
                hidden_states_len = (
                    hidden_states_len if hidden_states_len > 0 else hidden_states[0][0].shape[1]
                )
                print(f"hidden_states_len: {hidden_states_len}")
                # new generate thinker_token_embeds
                thinker_new_token_embeds = hidden_states[0][0][:, :hidden_states_len, :]
                hidden_states = (
                    (thinker_new_token_embeds,) + hidden_states[0][1:],
                ) + hidden_states[1:]
                # new generate thinker_hidden_states
                thinker_new_hidden_states = hidden_states[0][-1][:, :hidden_states_len, :]
                hidden_states = (
                    hidden_states[0][:-1] + (thinker_new_hidden_states,),
                ) + hidden_states[1:]

            # Decode and print only the text generated in this step
            step_new_text = processor.decode(step_new_ids[0], skip_special_tokens=True)
            yield {
                "thinker_generate_text": step_new_text,
                "thinker_generate_ids": step_new_ids,
                "thinker_generate_hidden_states": hidden_states,
            }  # TODO: put async queue here
            generated_text += step_new_text
            total_new_tokens_generated += num_step_new_tokens

            # Update the full sequence
            full_generated_ids = torch.cat([full_generated_ids, step_new_ids], dim=1)

            # Prepare for the next iteration:
            # Input is only the last generated token
            # NOTE: need use past_key_values to keep the context by manually,
            # current_input_ids = step_new_ids[:, -1:]
            # so we can't use the last generated token, use cache instead
            # input ids need to be the full sequence for next generation
            current_input_ids = full_generated_ids

            # Update past_key_values
            # past_key_values = outputs.past_key_values

            # Update attention mask by appending 1s for the new tokens
            full_attention_mask = torch.cat(
                [full_attention_mask, torch.ones_like(step_new_ids)], dim=1
            )
            current_attention_mask = full_attention_mask

            # torch.cuda.empty_cache()

            # Check if EOS token was generated in this step
            if step_new_ids[0, -1].item() in eos_token_ids:
                print("EOS token generated.")
                break

            # Check if max_new_tokens limit is reached (after processing the step)
            if total_new_tokens_generated >= max_new_tokens:
                print("Max new tokens limit reached.")
                break

        print(f"Total generated text: {generated_text}")
        print(f"Total new tokens generated: {total_new_tokens_generated}")
        print(
            f"max_tokens_per_step: {max_tokens_per_step} | first chunk generated cost: {times[0]} s | total cost: {sum(times)} s"
        )

    def talker_generate_chunk(
        input_ids,
        attention_mask,
        thinker_chunk_stream,
        speaker=DEFAULT_SPEAKER,
        talker_eos_token_id: list[int] = [8292, 8294],
    ):
        for chunk in thinker_chunk_stream:
            thinker_generate_text = chunk["thinker_generate_text"]
            if thinker_generate_text in " \n\r,;.?!，；。？！":
                yield (thinker_generate_text, torch.empty([1, 0]))
                continue
            thinker_generate_hidden_states = chunk["thinker_generate_hidden_states"]
            if thinker_generate_hidden_states is None:
                yield (thinker_generate_text, torch.empty([1, 0]))
                continue
            thinker_generate_ids = chunk["thinker_generate_ids"].to(model.talker.device)
            thinker_token_embeds = [
                token_hidden_states[0].to(model.talker.device)
                for token_hidden_states in thinker_generate_hidden_states
            ]
            thinker_hidden_states = [
                token_hidden_states[-1].to(model.talker.device)
                for token_hidden_states in thinker_generate_hidden_states
            ]
            print(
                f"[{thinker_generate_text}] len(thinker_generate_hidden_states):{len(thinker_generate_hidden_states)}"
            )
            for i in range(len(thinker_generate_hidden_states)):
                print(
                    f"[{thinker_generate_text}] thinker_generate_hidden_states[{i}]:{thinker_generate_hidden_states[i][0].shape}, {thinker_generate_hidden_states[i][-1].shape}"
                )
            # print(
            #    f"[{thinker_generate_text}] thinker_generate_hidden_states[0]:{thinker_generate_hidden_states[0][0][:,:5,:]}, {thinker_generate_hidden_states[0][-1][:,:5,:]}"
            # )

            talker_text_bos_token = model.speaker_map[speaker]["bos_token"]
            talker_input_text_ids = torch.cat(
                [
                    input_ids.to(model.talker.device),
                    torch.tensor(
                        [[talker_text_bos_token]], dtype=torch.long, device=model.talker.device
                    ),
                    thinker_generate_ids[:, :1],
                ],
                dim=-1,
            )
            print(
                f"[{thinker_generate_text}] talker_input_text_ids.shape:{talker_input_text_ids.shape}"
            )

            talker_input_ids = torch.cat(
                [
                    torch.full_like(
                        input_ids,
                        fill_value=model.talker.codec_mask_token,
                        device=model.talker.device,
                    ),
                    torch.tensor(
                        [[model.talker.codec_pad_token]],
                        dtype=torch.long,
                        device=model.talker.device,
                    ),
                    torch.tensor(
                        [[model.talker.codec_bos_token]],
                        dtype=torch.long,
                        device=model.talker.device,
                    ),
                ],
                dim=1,
            )
            print(f"[{thinker_generate_text}] talker_input_ids.shape:{talker_input_ids.shape}")

            thinker_embed_tokens = model.thinker.get_input_embeddings()
            thinker_reply_part = torch.cat(thinker_hidden_states[1:], dim=1) + torch.cat(
                thinker_token_embeds[1:], dim=1
            )
            talker_inputs_embeds = thinker_hidden_states[0] + thinker_token_embeds[0]
            talker_text_bos_token = torch.tensor(
                [[talker_text_bos_token]], dtype=torch.long, device=model.thinker.device
            )
            talker_text_bos_embed = thinker_embed_tokens(talker_text_bos_token).to(
                model.talker.device
            )
            print(
                f"[{thinker_generate_text}] talker_inputs_embeds.shape {talker_inputs_embeds.shape} talker_text_bos_embed.shape {talker_text_bos_embed.shape} thinker_reply_part.shape {thinker_reply_part.shape}"
            )
            talker_inputs_embeds = torch.cat(
                [
                    talker_inputs_embeds,
                    talker_text_bos_embed,
                    thinker_reply_part[:, :1, :],
                ],
                dim=1,
            )
            print(
                f"[{thinker_generate_text}] talker_inputs_embeds.shape {talker_inputs_embeds.shape} talker_text_bos_embed.shape {talker_text_bos_embed.shape}"
            )

            eos_embedding = thinker_embed_tokens(
                torch.tensor(
                    [[model.talker.text_eos_token]], dtype=torch.long, device=model.thinker.device
                )
            ).to(model.talker.device)

            pad_embedding = thinker_embed_tokens(
                torch.tensor(
                    [[model.talker.text_pad_token]], dtype=torch.long, device=model.thinker.device
                )
            ).to(model.talker.device)
            thinker_reply_part = torch.cat(
                [
                    thinker_reply_part[:, 1:, :],
                    eos_embedding,
                    pad_embedding,
                ],
                dim=1,
            )
            print(f"[{thinker_generate_text}] thinker_reply_part.shape:{thinker_reply_part.shape}")

            talker_attention_mask = None
            if attention_mask is not None:
                talker_attention_mask = torch.cat(
                    [attention_mask, attention_mask.new_ones((1, 2))], dim=1
                ).to(model.talker.device)

            streamer = TokenStreamer(skip_prompt=True)
            talker_kwargs = dict(
                input_ids=talker_input_ids,
                streamer=streamer,
                input_text_ids=talker_input_text_ids,
                thinker_reply_part=thinker_reply_part,
                inputs_embeds=talker_inputs_embeds,
                attention_mask=talker_attention_mask,
                suppress_tokens=[model.talker.codec_bos_token],
                eos_token_id=talker_eos_token_id,
                do_sample=True,
                top_k=10,
                top_p=0.9,
                temperature=0.95,
                repetition_penalty=1.1,
                min_new_tokens=0,
                max_new_tokens=8192,
            )
            # print(talker_kwargs.keys())
            thread = Thread(target=model.talker.generate, kwargs=talker_kwargs)
            thread.start()

            # 3. Generate wavs from code
            if model.token2wav.dtype != torch.float:
                model.token2wav.float()

            code2wav_times = []
            talker_generate_codes = []
            times = []
            start_time = perf_counter()
            pre_offset = 0
            for token_id in streamer:
                times.append(perf_counter() - start_time)
                start_time = perf_counter()
                if token_id in talker_eos_token_id:
                    break
                talker_generate_codes.append(token_id)
                chunk_code_length = len(talker_generate_codes) * 2 - 24
                if chunk_code_length > 0 and chunk_code_length % 48 == 0:
                    codes_tensor = torch.tensor(
                        [talker_generate_codes[pre_offset:]],
                        dtype=torch.long,
                        device=model.talker.device,
                    )
                    pre_offset = len(talker_generate_codes)
                    wav = (
                        model.token2wav(
                            codes_tensor.to(model.token2wav.device),
                            conditioning=model.speaker_map[speaker]["cond"]
                            .to(model.token2wav.device)
                            .float(),
                            reference_mel=model.speaker_map[speaker]["ref_mel"]
                            .to(model.token2wav.device)
                            .float(),
                            num_steps=10,
                            guidance_scale=0.5,
                            sway_coefficient=-1.0,
                        )
                        .unsqueeze(0)
                        .detach()
                    )
                    code2wav_times.append(perf_counter() - start_time)
                    yield (thinker_generate_text, wav)
                    start_time = perf_counter()

            print(
                f"[{thinker_generate_text}] generate first token cost time: {times[0]} s, {len(times)} tokens cost time: {sum(times)} s"
            )

            if len(talker_generate_codes) > pre_offset:
                codes_tensor = torch.tensor(
                    [talker_generate_codes[pre_offset:]],
                    dtype=torch.long,
                    device=model.talker.device,
                )
                wav = (
                    model.token2wav(
                        codes_tensor.to(model.token2wav.device),
                        conditioning=model.speaker_map[speaker]["cond"]
                        .to(model.token2wav.device)
                        .float(),
                        reference_mel=model.speaker_map[speaker]["ref_mel"]
                        .to(model.token2wav.device)
                        .float(),
                        num_steps=10,
                        guidance_scale=0.5,
                        sway_coefficient=-1.0,
                    )
                    .unsqueeze(0)
                    .detach()
                )
                code2wav_times.append(perf_counter() - start_time)
                yield (thinker_generate_text, wav)

            print(
                f"[{thinker_generate_text}] code2wav streaming first chunk time: {code2wav_times[0]} s | cost: {sum(code2wav_times)} s"
            )

            torch.cuda.empty_cache()

    def generate_stream(
        messages,
        use_audio_in_video=False,
        speaker=DEFAULT_SPEAKER,
        thinker_max_new_tokens=2048,
        thinker_max_tokens_per_step=10,  # Controls how many tokens to generate *per step*
        thinker_stop_strings_per_step=[".", "。"],
        thinker_eos_token_ids=[151644, 151645],  # Define EOS tokens
    ):
        print(messages)
        text = processor.apply_chat_template(messages, tokenize=False, add_generation_prompt=True)
        # image_inputs, video_inputs = process_vision_info([messages])
        audios, images, videos = process_mm_info(messages, use_audio_in_video=use_audio_in_video)
        print(text)
        {print(f"audios[{i}]: {item.shape}") for i, item in enumerate(audios)} if audios else print(
            audios
        )
        {print(f"images[{i}]: {item}") for i, item in enumerate(images)} if images else print(
            images
        )
        {print(f"videos[{i}]: {item.shape}") for i, item in enumerate(videos)} if videos else print(
            videos
        )

        inputs = processor(
            text=text,
            audio=audios,
            images=images,
            videos=videos,
            return_tensors="pt",
            padding=True,
            use_audio_in_video=use_audio_in_video,
        )
        inputs = inputs.to(model.device).to(model.dtype)
        for k, v in inputs.items():
            print(k, v.shape)
        thinker_chunk_stream = thinker_generate_chunk(
            inputs,
            max_new_tokens=thinker_max_new_tokens,
            max_tokens_per_step=thinker_max_tokens_per_step,
            stop_strings_per_step=thinker_stop_strings_per_step,
            use_audio_in_video=use_audio_in_video,
            eos_token_ids=thinker_eos_token_ids,
            output_hidden_states=True,
        )
        return talker_generate_chunk(
            input_ids=inputs["input_ids"],
            attention_mask=inputs["attention_mask"],
            thinker_chunk_stream=thinker_chunk_stream,
            speaker=speaker,
        )

    @torch.no_grad()
    def thinker_talker_inference_stream(
        messages,
        use_audio_in_video=False,
        speaker=DEFAULT_SPEAKER,
        talker_eos_token_id: list[int] = [8292, 8294],
    ):
        print(messages)
        text = processor.apply_chat_template(messages, tokenize=False, add_generation_prompt=True)
        # image_inputs, video_inputs = process_vision_info([messages])
        audios, images, videos = process_mm_info(messages, use_audio_in_video=use_audio_in_video)
        print(text)
        {print(f"audios[{i}]: {item.shape}") for i, item in enumerate(audios)} if audios else print(
            audios
        )
        {print(f"images[{i}]: {item}") for i, item in enumerate(images)} if images else print(
            images
        )
        {print(f"videos[{i}]: {item.shape}") for i, item in enumerate(videos)} if videos else print(
            videos
        )
        inputs = processor(
            text=text,
            audio=audios,
            images=images,
            videos=videos,
            return_tensors="pt",
            padding=True,
            use_audio_in_video=use_audio_in_video,
        )
        inputs = inputs.to(model.device).to(model.dtype)

        thinker_result = model.thinker.generate(
            **inputs,
            use_audio_in_video=use_audio_in_video,
            do_sample=True,
            top_k=10,
            top_p=0.9,
            temperature=0.95,
            repetition_penalty=1.1,
            min_new_tokens=0,
            max_new_tokens=2048,
            output_hidden_states=True,
            return_dict_in_generate=True,
        )
        print(f" len(thinker_generate_hidden_states):{len(thinker_result.hidden_states)}")
        for i in range(len(thinker_result.hidden_states)):
            print(
                f"thinker_generate_hidden_states[{i}]:{thinker_result.hidden_states[i][0].shape}, {thinker_result.hidden_states[i][-1].shape}"
            )
        # 2. Generate speech tokens from talker module
        input_ids = inputs["input_ids"]
        thinker_generate_ids = thinker_result.sequences[:, input_ids.size(1) :].to(
            model.talker.device
        )
        thinker_token_embeds = [
            token_hidden_states[0].to(model.talker.device)
            for token_hidden_states in thinker_result.hidden_states
        ]
        thinker_hidden_states = [
            token_hidden_states[-1].to(model.talker.device)
            for token_hidden_states in thinker_result.hidden_states
        ]
        gen_text = processor.batch_decode(
            thinker_result.sequences,
            skip_special_tokens=True,
            clean_up_tokenization_spaces=False,
        )
        # print(gen_text)

        talker_text_bos_token = model.speaker_map[speaker]["bos_token"]
        talker_input_text_ids = torch.cat(
            [
                input_ids.to(model.talker.device),
                torch.tensor(
                    [[talker_text_bos_token]], dtype=torch.long, device=model.talker.device
                ),
                thinker_generate_ids[:, :1],
            ],
            dim=-1,
        )
        print(f"talker_input_text_ids.shape:{talker_input_text_ids.shape}")

        talker_input_ids = torch.cat(
            [
                torch.full_like(
                    input_ids, fill_value=model.talker.codec_mask_token, device=model.talker.device
                ),
                torch.tensor(
                    [[model.talker.codec_pad_token]], dtype=torch.long, device=model.talker.device
                ),
                torch.tensor(
                    [[model.talker.codec_bos_token]], dtype=torch.long, device=model.talker.device
                ),
            ],
            dim=1,
        )
        print(f"talker_input_ids.shape:{talker_input_ids.shape}")

        thinker_embed_tokens = model.thinker.get_input_embeddings()
        thinker_reply_part = torch.cat(thinker_hidden_states[1:], dim=1) + torch.cat(
            thinker_token_embeds[1:], dim=1
        )
        talker_inputs_embeds = thinker_hidden_states[0] + thinker_token_embeds[0]
        talker_text_bos_token = torch.tensor(
            [[talker_text_bos_token]], dtype=torch.long, device=model.thinker.device
        )
        talker_text_bos_embed = thinker_embed_tokens(talker_text_bos_token).to(model.talker.device)
        print(
            f"talker_inputs_embeds.shape {talker_inputs_embeds.shape} talker_text_bos_embed.shape {talker_text_bos_embed.shape} thinker_reply_part.shape {thinker_reply_part.shape}"
        )
        talker_inputs_embeds = torch.cat(
            [
                talker_inputs_embeds,
                talker_text_bos_embed,
                thinker_reply_part[:, :1, :],
            ],
            dim=1,
        )
        print(
            f"talker_inputs_embeds.shape {talker_inputs_embeds.shape} talker_text_bos_embed.shape {talker_text_bos_embed.shape}"
        )

        eos_embedding = thinker_embed_tokens(
            torch.tensor(
                [[model.talker.text_eos_token]], dtype=torch.long, device=model.thinker.device
            )
        ).to(model.talker.device)

        pad_embedding = thinker_embed_tokens(
            torch.tensor(
                [[model.talker.text_pad_token]], dtype=torch.long, device=model.thinker.device
            )
        ).to(model.talker.device)

        thinker_reply_part = torch.cat(
            [
                thinker_reply_part[:, 1:, :],
                eos_embedding,
                pad_embedding,
            ],
            dim=1,
        )
        print(f"thinker_reply_part.shape:{thinker_reply_part.shape}")

        talker_attention_mask = None
        if "attention_mask" in inputs:
            talker_attention_mask = torch.cat(
                [inputs["attention_mask"], inputs["attention_mask"].new_ones((1, 2))], dim=1
            ).to(model.talker.device)

        # talker_result = model.talker.generate(
        #    input_ids=talker_input_ids,
        #    input_text_ids=talker_input_text_ids,
        #    thinker_reply_part=thinker_reply_part,
        #    inputs_embeds=talker_inputs_embeds,
        #    attention_mask=talker_attention_mask,
        #    suppress_tokens=[model.talker.codec_bos_token],
        #    do_sample=True,
        #    top_k=10,
        #    top_p=0.9,
        #    temperature=0.95,
        #    repetition_penalty=1.1,
        #    min_new_tokens=0,
        #    max_new_tokens=8192,
        # )
        # talker_generate_codes = talker_result[:, talker_input_ids.shape[1] : -1]
        # print(talker_generate_codes)

        streamer = TokenStreamer(skip_prompt=True)
        talker_kwargs = dict(
            input_ids=talker_input_ids,
            streamer=streamer,
            input_text_ids=talker_input_text_ids,
            thinker_reply_part=thinker_reply_part,
            inputs_embeds=talker_inputs_embeds,
            attention_mask=talker_attention_mask,
            suppress_tokens=[model.talker.codec_bos_token],
            eos_token_id=talker_eos_token_id,
            do_sample=True,
            top_k=10,
            top_p=0.9,
            temperature=0.95,
            repetition_penalty=1.1,
            min_new_tokens=0,
            max_new_tokens=8192,
        )
        # print(talker_kwargs.keys())
        thread = Thread(target=model.talker.generate, kwargs=talker_kwargs)
        thread.start()

        # 3. Generate wavs from code
        if model.token2wav.dtype != torch.float:
            model.token2wav.float()

        code2wav_times = []
        talker_generate_codes = []
        times = []
        start_time = perf_counter()
        pre_offset = 0
        for token_id in streamer:
            times.append(perf_counter() - start_time)
            start_time = perf_counter()
            if token_id in talker_eos_token_id:
                break
            talker_generate_codes.append(token_id)
            chunk_code_length = len(talker_generate_codes) * 2 - 24
            if chunk_code_length > 0 and chunk_code_length % 48 == 0:
                codes_tensor = torch.tensor(
                    [talker_generate_codes[pre_offset:]],
                    dtype=torch.long,
                    device=model.talker.device,
                )
                pre_offset = len(talker_generate_codes)
                wav = (
                    model.token2wav(
                        codes_tensor.to(model.token2wav.device),
                        conditioning=model.speaker_map[speaker]["cond"]
                        .to(model.token2wav.device)
                        .float(),
                        reference_mel=model.speaker_map[speaker]["ref_mel"]
                        .to(model.token2wav.device)
                        .float(),
                        num_steps=10,
                        guidance_scale=0.5,
                        sway_coefficient=-1.0,
                    )
                    .unsqueeze(0)
                    .detach()
                )
                code2wav_times.append(perf_counter() - start_time)
                yield (gen_text, wav)
                start_time = perf_counter()

        print(
            f"generate first token cost time: {times[0]} s, {len(times)} tokens cost time: {sum(times)} s"
        )

        if len(talker_generate_codes) > pre_offset:
            codes_tensor = torch.tensor(
                [talker_generate_codes[pre_offset:]],
                dtype=torch.long,
                device=model.talker.device,
            )
            wav = (
                model.token2wav(
                    codes_tensor.to(model.token2wav.device),
                    conditioning=model.speaker_map[speaker]["cond"]
                    .to(model.token2wav.device)
                    .float(),
                    reference_mel=model.speaker_map[speaker]["ref_mel"]
                    .to(model.token2wav.device)
                    .float(),
                    num_steps=10,
                    guidance_scale=0.5,
                    sway_coefficient=-1.0,
                )
                .unsqueeze(0)
                .detach()
            )
            code2wav_times.append(perf_counter() - start_time)
            yield (gen_text, wav)

        print(
            f"code2wav streaming first chunk time: {code2wav_times[0]} s | cost: {sum(code2wav_times)} s"
        )

        torch.cuda.empty_cache()


@app.function(
    gpu=os.getenv("IMAGE_GPU", "L40s"),
    cpu=2.0,
    retries=0,
    image=omni_img,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        ASSETS_DIR: assets_dir,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=1,
)
def run(func):
    func()


def voice_chatting():
    import torchaudio

    sys_msg = {
        "role": "system",
        "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}],
    }

    for audio_path in ["guess_age_gender.wav", "translate_to_chinese.wav"]:
        audio_path = os.path.join(ASSETS_DIR, audio_path)
        audio_msg = [
            sys_msg,
            {
                "role": "user",
                "content": [
                    {"type": "audio", "audio": audio_path},
                ],
            },
        ]

        texts, audio = inference(audio_msg, return_audio=True, use_audio_in_video=True)
        print(texts[0], audio.shape)

        save_audio_path = os.path.join(ASSETS_DIR, f"generated_{os.path.basename(audio_path)}")
        torchaudio.save(save_audio_path, audio, sample_rate=24000)
        print(f"Audio saved to {save_audio_path}")


def multi_round_omni_chatting():
    import torchaudio

    conversations = [
        {
            "role": "system",
            "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}],
        },
    ]
    for video_path in ["draw1.mp4", "draw2.mp4", "draw3.mp4"]:
        video_path = os.path.join(ASSETS_DIR, video_path)
        conversations.append({"role": "user", "content": [{"type": "video", "video": video_path}]})
        texts, audio = inference(conversations, return_audio=True, use_audio_in_video=True)
        print(texts[0], audio.shape)
        save_audio_path = os.path.join(ASSETS_DIR, f"generated_{os.path.basename(video_path)}")
        torchaudio.save(save_audio_path, audio, sample_rate=24000)
        print(f"Audio saved to {save_audio_path}")


def omni_chatting_for_math():
    import torchaudio

    video_path = os.path.join(ASSETS_DIR, "math.mp4")
    messages = [
        {
            "role": "system",
            "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}],
        },
        {
            "role": "user",
            "content": [
                {"type": "video", "video": video_path},
            ],
        },
    ]

    response, audio = inference(messages, return_audio=True, use_audio_in_video=True)
    print(response[0], audio.shape)

    save_audio_path = os.path.join(ASSETS_DIR, f"generated_{os.path.basename(video_path)}")
    torchaudio.save(save_audio_path, audio, sample_rate=24000)
    print(f"Audio saved to {save_audio_path}")


def omni_chatting_for_math_stream():
    import torchaudio
    import soundfile as sf

    video_path = os.path.join(ASSETS_DIR, "math.mp4")
    messages = [
        {
            "role": "system",
            "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}],
        },
        {
            "role": "user",
            "content": [
                {"type": "video", "video": video_path},
            ],
        },
    ]

    for _ in range(1):  # warmup and test
        streamer = thinker_talker_inference_stream(messages, use_audio_in_video=True)
        audios = []
        times = []
        start_time = time.perf_counter()
        for i, (texts, audio) in enumerate(streamer):
            if i == 0:
                print(texts[0])
            times.append(time.perf_counter() - start_time)
            audios.append(audio.squeeze().cpu().numpy())
            # save_audio_path = os.path.join(ASSETS_DIR, f"omni_chatting_for_math_stream_{i}.wav")
            # torchaudio.save(save_audio_path, audio, sample_rate=24000)
            # print(f"Audio saved to {save_audio_path}")
            start_time = time.perf_counter()

        save_audio_path = os.path.join(ASSETS_DIR, f"omni_chatting_for_math_stream.wav")
        sf.write(save_audio_path, np.concatenate(audios), samplerate=24000)
        print(f"Audio saved to {save_audio_path}")
        info = sf.info(save_audio_path, verbose=True)
        print(
            f"thinker->talker->code2wav streaming first chunk time: {times[0]} s | wav duration: {info.duration} s | cost: {sum(times)} s | RTF: {sum(times)/info.duration}"
        )


def omni_chatting_for_math_chunk_stream():
    import torchaudio
    import soundfile as sf

    video_path = os.path.join(ASSETS_DIR, "math.mp4")
    messages = [
        {
            "role": "system",
            "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}],
        },
        {
            "role": "user",
            "content": [
                {"type": "video", "video": video_path},
            ],
        },
    ]

    thinker_eos_token_ids = [151644, 151645]
    print(thinker_eos_token_ids)
    for _ in range(1):  # warmup and test
        streamer = generate_stream(
            messages,
            use_audio_in_video=True,
            thinker_max_new_tokens=100,
            thinker_max_tokens_per_step=15,
            thinker_stop_strings_per_step=[".", "。"],
            thinker_eos_token_ids=thinker_eos_token_ids,
        )
        gen_text = ""
        gen_all_text = ""
        audios = []
        times = []
        start_time = time.perf_counter()
        for i, (text, wav) in enumerate(streamer):
            times.append(time.perf_counter() - start_time)
            if gen_text != text:
                gen_text = text
                gen_all_text += gen_text
            print(text, wav.shape)
            audios.append(wav.squeeze().cpu().numpy())
            # save_audio_path = os.path.join(ASSETS_DIR, f"omni_chatting_for_math_chunk_stream-{i}-{text}.wav")
            # torchaudio.save(save_audio_path, wav, sample_rate=24000)
            # print(f"Audio saved to {save_audio_path}")
            start_time = time.perf_counter()

        print(f"gen all text: {gen_all_text}")
        save_audio_path = os.path.join(ASSETS_DIR, f"omni_chatting_for_math_chunk_stream.wav")
        sf.write(save_audio_path, np.concatenate(audios), samplerate=24000)
        print(f"All Audio saved to {save_audio_path}")
        info = sf.info(save_audio_path, verbose=True)
        print(
            f"thinker->talker->code2wav chunk streaming first chunk time: {times[0]} s | wav duration: {info.duration} s | cost: {sum(times)} s | RTF: {sum(times)/info.duration}"
        )


def omni_chatting_for_music():
    import torchaudio

    video_path = os.path.join(ASSETS_DIR, "music.mp4")
    messages = [
        {
            "role": "system",
            "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}],
        },
        {
            "role": "user",
            "content": [
                {"type": "video", "video": video_path},
            ],
        },
    ]
    response, audio = inference(messages, return_audio=True, use_audio_in_video=True)
    print(response[0])

    save_audio_path = os.path.join(ASSETS_DIR, f"generated_{os.path.basename(video_path)}")
    torchaudio.save(save_audio_path, audio, sample_rate=24000)
    print(f"Audio saved to {save_audio_path}")


def omni_chatting_for_music_stream():
    import torchaudio
    import soundfile as sf

    video_path = os.path.join(ASSETS_DIR, "music.mp4")
    messages = [
        {
            "role": "system",
            "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}],
        },
        {
            "role": "user",
            "content": [
                {"type": "video", "video": video_path},
            ],
        },
    ]
    for _ in range(1):  # warmup and test
        streamer = thinker_talker_inference_stream(messages, use_audio_in_video=True)
        audios = []
        times = []
        start_time = time.perf_counter()
        for i, (texts, audio) in enumerate(streamer):
            if i == 0:
                print(texts[0])
            times.append(time.perf_counter() - start_time)
            audios.append(audio.squeeze().cpu().numpy())
            # save_audio_path = os.path.join(ASSETS_DIR, f"omni_chatting_for_math_stream_{i}.wav")
            # torchaudio.save(save_audio_path, audio, sample_rate=24000)
            # print(f"Audio saved to {save_audio_path}")
            start_time = time.perf_counter()

        save_audio_path = os.path.join(ASSETS_DIR, f"omni_chatting_for_music_stream.wav")
        sf.write(save_audio_path, np.concatenate(audios), samplerate=24000)
        print(f"Audio saved to {save_audio_path}")
        info = sf.info(save_audio_path, verbose=True)
        print(
            f"thinker->talker->code2wav streaming first chunk time: {times[0]} s | wav duration: {info.duration} s | cost: {sum(times)} s | RTF: {sum(times)/info.duration}"
        )


def omni_chatting_for_music_chunk_stream():
    import torchaudio
    import soundfile as sf

    video_path = os.path.join(ASSETS_DIR, "music.mp4")
    messages = [
        {
            "role": "system",
            "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}],
        },
        {
            "role": "user",
            "content": [
                {"type": "video", "video": video_path},
            ],
        },
    ]
    thinker_eos_token_ids = [151644, 151645]
    print(thinker_eos_token_ids)
    for _ in range(1):  # warmup and test
        streamer = generate_stream(
            messages,
            use_audio_in_video=True,
            thinker_max_new_tokens=100,
            thinker_max_tokens_per_step=15,
            thinker_stop_strings_per_step=[".", "。"],
            thinker_eos_token_ids=thinker_eos_token_ids,
        )
        gen_text = ""
        gen_all_text = ""
        audios = []
        times = []
        start_time = time.perf_counter()
        for i, (text, wav) in enumerate(streamer):
            times.append(time.perf_counter() - start_time)
            if gen_text != text:
                gen_text = text
                gen_all_text += gen_text
            print(text, wav.shape)
            audios.append(wav.squeeze().cpu().numpy())
            # save_audio_path = os.path.join(ASSETS_DIR, f"omni_chatting_for_music_chunk_stream-{i}-{text}.wav")
            # torchaudio.save(save_audio_path, wav, sample_rate=24000)
            # print(f"Audio saved to {save_audio_path}")
            start_time = time.perf_counter()

        print(f"gen all text: {gen_all_text}")
        save_audio_path = os.path.join(ASSETS_DIR, f"omni_chatting_for_music_chunk_stream.wav")
        sf.write(save_audio_path, np.concatenate(audios), samplerate=24000)
        print(f"All Audio saved to {save_audio_path}")
        info = sf.info(save_audio_path, verbose=True)
        print(
            f"thinker->talker->code2wav chunk streaming first chunk time: {times[0]} s | wav duration: {info.duration} s | cost: {sum(times)} s | RTF: {sum(times)/info.duration}"
        )


def screen_recording_interaction():
    video_path = os.path.join(ASSETS_DIR, "screen.mp4")
    for prompt in [
        "What the browser is used in this video?",
        "浏览器中的论文叫什么名字?",
        "这篇论文主要解决什么问题呢？",
    ]:
        messages = [
            {
                "role": "system",
                "content": [{"type": "text", "text": "You are a helpful assistant."}],
            },
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": prompt},
                    {"type": "video", "video": video_path},
                ],
            },
        ]

        response, _ = inference(messages, return_audio=False, use_audio_in_video=False)
        print(response[0])


def screen_recording_interaction_stream():
    video_path = os.path.join(ASSETS_DIR, "screen.mp4")
    for prompt in [
        "What the browser is used in this video?",
        "浏览器中的论文叫什么名字?",
        "这篇论文主要解决什么问题呢？",
    ]:
        messages = [
            {
                "role": "system",
                "content": [{"type": "text", "text": "You are a helpful assistant."}],
            },
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": prompt},
                    {"type": "video", "video": video_path},
                ],
            },
        ]

        text_stream = thinker_inference_stream(messages, use_audio_in_video=False)
        for text in text_stream:
            print(text)


def screen_recording_interaction_chunk_stream():
    video_path = os.path.join(ASSETS_DIR, "screen.mp4")
    for prompt in [
        "What the browser is used in this video?",
        "浏览器中的论文叫什么名字?",
        "这篇论文主要解决什么问题呢？",
    ]:
        messages = [
            {
                "role": "system",
                "content": [{"type": "text", "text": "You are a helpful assistant."}],
            },
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": prompt},
                    {"type": "video", "video": video_path},
                ],
            },
        ]

        text_stream = thinker_inference_chunk_stream(messages, use_audio_in_video=False)
        for text in text_stream:
            print(text)


def universal_audio_understanding():
    for case in [
        {
            "audio_path": "1272-128104-0000.flac",
            "prompt": "Transcribe the English audio into text without any punctuation marks.",
            "sys_prompt": "You are a speech recognition model.",
        },
        {
            "audio_path": "BAC009S0764W0121.wav",
            "prompt": "请将这段中文语音转换为纯文本，去掉标点符号。",
            "sys_prompt": "You are a speech recognition model.",
        },
        {
            "audio_path": "10000611681338527501.wav",
            "prompt": "Transcribe the Russian audio into text without including any punctuation marks.",
            "sys_prompt": "You are a speech recognition model.",
        },
        {
            "audio_path": "7105431834829365765.wav",
            "prompt": "Transcribe the French audio into text without including any punctuation marks.",
            "sys_prompt": "You are a speech recognition model.",
        },
        {
            "audio_path": "1272-128104-0000.flac",
            "prompt": "Listen to the provided English speech and produce a translation in Chinese text.",
            "sys_prompt": "You are a speech translation model.",
        },
        {
            "audio_path": "cough.wav",
            "prompt": "Classify the given human vocal sound in English.",
            "sys_prompt": "You are a voice classification model.",
        },
    ]:
        audio_path = os.path.join(ASSETS_DIR, case["audio_path"])
        messages = [
            {"role": "system", "content": [{"type": "text", "text": case["sys_prompt"]}]},
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": case["prompt"]},
                    {"type": "audio", "audio": audio_path},
                ],
            },
        ]
        texts, _ = inference(messages, use_audio_in_video=True, return_audio=False)
        print(texts[0])


def video_information_extracting():
    video_path = os.path.join(ASSETS_DIR, "shopping.mp4")
    sys_msg = {
        "role": "system",
        "content": [{"type": "text", "text": "You are a helpful assistant."}],
    }
    for prompt in [
        "How many kind of drinks can you see in the video?",
        "How many bottles of drinks have I picked up?",
        "How many milliliters are there in the bottle I picked up second time?",
        "视屏中的饮料叫什么名字呢？",
        "跑步🏃🏻累了，适合喝什么饮料补充体力呢？",
    ]:
        messages = [
            sys_msg,
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": prompt},
                    {"type": "video", "video": video_path},
                ],
            },
        ]
        texts, _ = inference(messages, return_audio=False, use_audio_in_video=False)
        print(texts[0])


def video_information_extracting_stream():
    video_path = os.path.join(ASSETS_DIR, "shopping.mp4")
    sys_msg = {
        "role": "system",
        "content": [{"type": "text", "text": "You are a helpful assistant."}],
    }
    for prompt in [
        "How many kind of drinks can you see in the video?",
        "How many bottles of drinks have I picked up?",
        "How many milliliters are there in the bottle I picked up second time?",
        "视屏中的饮料叫什么名字呢？",
        "跑步🏃🏻累了，适合喝什么饮料补充体力呢？",
    ]:
        messages = [
            sys_msg,
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": prompt},
                    {"type": "video", "video": video_path},
                ],
            },
        ]
        text_stream = thinker_inference_stream(messages, use_audio_in_video=False)
        for text in text_stream:
            print(text)


def video_information_extracting_chunk_stream():
    video_path = os.path.join(ASSETS_DIR, "shopping.mp4")
    sys_msg = {
        "role": "system",
        "content": [{"type": "text", "text": "You are a helpful assistant."}],
    }
    for prompt in [
        "How many kind of drinks can you see in the video?",
        "How many bottles of drinks have I picked up?",
        "How many milliliters are there in the bottle I picked up second time?",
        "视屏中的饮料叫什么名字呢？",
        "跑步🏃🏻累了，适合喝什么饮料补充体力呢？",
    ]:
        messages = [
            sys_msg,
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": prompt},
                    {"type": "video", "video": video_path},
                ],
            },
        ]
        text_stream = thinker_inference_chunk_stream(messages, use_audio_in_video=False)
        for text in text_stream:
            print(text)


def batch_requests():
    """need return_audio=False"""
    # Conversation with video only
    conversation1 = [
        {"role": "system", "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}]},
        {
            "role": "user",
            "content": [
                {"type": "video", "video": os.path.join(ASSETS_DIR, "draw1.mp4")},
            ],
        },
    ]

    # Conversation with audio only
    conversation2 = [
        {"role": "system", "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}]},
        {
            "role": "user",
            "content": [
                {"type": "audio", "audio": os.path.join(ASSETS_DIR, "1272-128104-0000.flac")},
            ],
        },
    ]

    # Conversation with pure text
    conversation3 = [
        {"role": "system", "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}]},
        {"role": "user", "content": [{"type": "text", "text": "who are you?"}]},
    ]

    # Conversation with mixed media
    conversation4 = [
        {"role": "system", "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}]},
        {
            "role": "user",
            "content": [
                {"type": "image", "image": os.path.join(ASSETS_DIR, "03-Confusing-Pictures.jpg")},
                {"type": "video", "video": os.path.join(ASSETS_DIR, "music.mp4")},
                {"type": "audio", "audio": os.path.join(ASSETS_DIR, "1272-128104-0000.flac")},
                {
                    "type": "text",
                    "text": "What are the elements can you see and hear in these medias?",
                },
            ],
        },
    ]

    # Combine messages for batch processing
    conversations = [conversation1, conversation2, conversation3, conversation4]
    texts, _ = inference(conversations, return_audio=False, use_audio_in_video=True)
    print(texts)


def image_stream():
    for case in [
        {
            "image_path": "03-Confusing-Pictures.jpg",
            "prompt": "请描述一下图片中的内容",
            "sys_prompt": "You are a vision recognition model.",
        },
    ]:
        image_path = os.path.join(ASSETS_DIR, case["image_path"])
        messages = [
            {"role": "system", "content": [{"type": "text", "text": case["sys_prompt"]}]},
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": case["prompt"]},
                    {"type": "image", "image": image_path},
                ],
            },
        ]
        text_streamer = thinker_inference_stream(messages, use_audio_in_video=False)
        for text in text_streamer:
            print(text)


def image_chunk_stream():
    for case in [
        {
            "image_path": "03-Confusing-Pictures.jpg",
            "prompt": "请描述一下图片中的内容",
            "sys_prompt": "You are a vision recognition model.",
        },
    ]:
        image_path = os.path.join(ASSETS_DIR, case["image_path"])
        messages = [
            {"role": "system", "content": [{"type": "text", "text": case["sys_prompt"]}]},
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": case["prompt"]},
                    {"type": "image", "image": image_path},
                ],
            },
        ]
        text_streamer = thinker_inference_chunk_stream(
            messages,
            use_audio_in_video=False,
            output_hidden_states=False,
            max_new_tokens=1024,
        )
        for text in text_streamer:
            print(text)


def asr_stream():
    for case in [
        {
            "audio_path": "1272-128104-0000.flac",
            "prompt": "Listen to the provided English speech and produce a translation in Chinese text.",
            "sys_prompt": "You are a speech translation model.",
        },
        {
            "audio_path": "BAC009S0764W0121.wav",
            "prompt": "请将这段中文语音转换为纯文本",
            "sys_prompt": "You are a speech recognition model.",
        },
        {
            "audio_path": "asr_example_zh.wav",
            "prompt": "请将这段中文语音转换为纯文本",
            "sys_prompt": "You are a speech recognition model.",
        },
    ]:
        audio_path = os.path.join(ASSETS_DIR, case["audio_path"])
        messages = [
            {"role": "system", "content": [{"type": "text", "text": case["sys_prompt"]}]},
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": case["prompt"]},
                    {"type": "audio", "audio": audio_path},
                ],
            },
        ]
        text_streamer = thinker_inference_stream(messages, use_audio_in_video=True)
        for text in text_streamer:
            print(text)


def asr_chunk_stream():
    for case in [
        {
            "audio_path": "asr_example_zh.wav",
            "prompt": "请将这段中文语音转换为纯文本",
            "sys_prompt": "You are a speech recognition model.",
        },
    ]:
        audio_path = os.path.join(ASSETS_DIR, case["audio_path"])
        messages = [
            {"role": "system", "content": [{"type": "text", "text": case["sys_prompt"]}]},
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": case["prompt"]},
                    {"type": "audio", "audio": audio_path},
                ],
            },
        ]
        chunk_streamer = thinker_inference_chunk_stream(messages, use_audio_in_video=False)
        for chunk in chunk_streamer:
            print(chunk)


def thinker_stream():
    messages = [
        {"role": "system", "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}]},
        {"role": "user", "content": [{"type": "text", "text": "who are you?"}]},
    ]
    chunk_stream = thinker_inference_stream(messages)
    for chunk in chunk_stream:
        print(chunk)


def thinker_chunk_stream():
    messages = [
        {"role": "system", "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}]},
        {"role": "user", "content": [{"type": "text", "text": "who are you?"}]},
    ]
    chunk_stream = thinker_inference_chunk_stream(
        messages,
        use_audio_in_video=False,
        output_hidden_states=False,
        max_new_tokens=100,
    )
    for chunk in chunk_stream:
        print(chunk)


def omni_chatting_stream():
    import torchaudio
    import soundfile as sf

    messages = [
        {"role": "system", "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}]},
        {"role": "user", "content": [{"type": "text", "text": "who are you?"}]},
    ]
    # response, audio = inference(
    #    messages, return_audio=True, use_audio_in_video=False, thinker_do_sample=True
    # )
    # print(response[0])
    # save_audio_path = os.path.join(ASSETS_DIR, f"generated_omni_chatting_stream.wav")
    # torchaudio.save(save_audio_path, audio, sample_rate=24000)
    # print(f"Audio saved to {save_audio_path}")
    # return

    for _ in range(1):  # warmup and test
        streamer = thinker_talker_inference_stream(messages, use_audio_in_video=False)
        audios = []
        times = []
        start_time = time.perf_counter()
        for i, (texts, audio) in enumerate(streamer):
            if i == 0:
                print(texts[0])
            times.append(time.perf_counter() - start_time)
            audios.append(audio.squeeze().cpu().numpy())
            # save_audio_path = os.path.join(ASSETS_DIR, f"generated_omni_chatting_stream_{i}.wav")
            # torchaudio.save(save_audio_path, audio, sample_rate=24000)
            # print(f"Audio saved to {save_audio_path}")
            start_time = time.perf_counter()

        save_audio_path = os.path.join(ASSETS_DIR, f"generated_omni_chatting_stream.wav")
        sf.write(save_audio_path, np.concatenate(audios), samplerate=24000)
        info = sf.info(save_audio_path, verbose=True)
        print(
            f"thinker->talker->code2wav streaming first chunk time: {times[0]} s | wav duration: {info.duration} s | cost: {sum(times)} s | RTF: {sum(times)/info.duration}"
        )


def omni_chatting_segment_stream():
    import torchaudio
    import soundfile as sf

    messages = [
        {"role": "system", "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}]},
        {"role": "user", "content": [{"type": "text", "text": "who are you?"}]},
    ]
    thinker_eos_token_ids = [151644, 151645]
    print(thinker_eos_token_ids)
    chunk_stream = generate_stream(
        messages,
        use_audio_in_video=False,
        thinker_max_new_tokens=100,
        thinker_max_tokens_per_step=15,
        thinker_stop_strings_per_step=[".", "。"],
        thinker_eos_token_ids=thinker_eos_token_ids,
    )

    gen_text = ""
    gen_all_text = ""
    audios = []
    times = []
    start_time = time.perf_counter()
    for i, (text, wav) in enumerate(chunk_stream):
        times.append(time.perf_counter() - start_time)
        if gen_text != text:
            gen_text = text
            gen_all_text += gen_text
        print(text, wav.shape)
        audios.append(wav.squeeze().cpu().numpy())
        save_audio_path = os.path.join(ASSETS_DIR, f"omni_chatting_segment_stream-{i}-{text}.wav")
        # torchaudio.save(save_audio_path, wav, sample_rate=24000)
        # print(f"Audio saved to {save_audio_path}")
        start_time = time.perf_counter()

    print(f"gen all text: {gen_all_text}")
    save_audio_path = os.path.join(ASSETS_DIR, f"omni_chatting_segment_stream.wav")
    sf.write(save_audio_path, np.concatenate(audios), samplerate=24000)
    print(f"All Audio saved to {save_audio_path}")
    info = sf.info(save_audio_path, verbose=True)
    print(
        f"thinker->talker->code2wav chunk streaming first chunk time: {times[0]} s | wav duration: {info.duration} s | cost: {sum(times)} s | RTF: {sum(times)/info.duration}"
    )


def image_chatting_stream():
    import torchaudio
    import soundfile as sf

    messages = [
        {"role": "system", "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}]},
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "请描述一下图片中的内容"},
                {"type": "image", "image": os.path.join(ASSETS_DIR, "03-Confusing-Pictures.jpg")},
            ],
        },
    ]
    # response, audio = inference(
    #    messages, return_audio=True, use_audio_in_video=False, thinker_do_sample=True
    # )
    # print(response[0])
    # save_audio_path = os.path.join(ASSETS_DIR, f"generated_image_chatting_stream.wav")
    # torchaudio.save(save_audio_path, audio, sample_rate=24000)
    # print(f"Audio saved to {save_audio_path}")
    # return

    for _ in range(1):  # warmup and test
        streamer = thinker_talker_inference_stream(messages, use_audio_in_video=False)
        audios = []
        times = []
        start_time = time.perf_counter()
        for i, (texts, audio) in enumerate(streamer):
            if i == 0:
                print(texts[0])
            times.append(time.perf_counter() - start_time)
            audios.append(audio.squeeze().cpu().numpy())
            # save_audio_path = os.path.join(ASSETS_DIR, f"generated_image_chatting_stream_{i}.wav")
            # torchaudio.save(save_audio_path, audio, sample_rate=24000)
            # print(f"Audio saved to {save_audio_path}")
            start_time = time.perf_counter()

        save_audio_path = os.path.join(ASSETS_DIR, f"generated_image_chatting_stream.wav")
        sf.write(save_audio_path, np.concatenate(audios), samplerate=24000)
        info = sf.info(save_audio_path, verbose=True)
        print(
            f"thinker->talker->code2wav streaming first chunk time: {times[0]} s | wav duration: {info.duration} s | cost: {sum(times)} s | RTF: {sum(times)/info.duration}"
        )


def image_chatting_segment_stream():
    import torchaudio
    import soundfile as sf

    messages = [
        {"role": "system", "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}]},
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "请描述一下图片中的内容"},
                {"type": "image", "image": os.path.join(ASSETS_DIR, "03-Confusing-Pictures.jpg")},
            ],
        },
    ]
    thinker_eos_token_ids = [151644, 151645]
    print(thinker_eos_token_ids)
    chunk_stream = generate_stream(
        messages,
        use_audio_in_video=False,
        thinker_max_new_tokens=100,
        thinker_max_tokens_per_step=15,
        thinker_stop_strings_per_step=[".", "。"],
        thinker_eos_token_ids=thinker_eos_token_ids,
    )

    gen_text = ""
    gen_all_text = ""
    audios = []
    times = []
    start_time = time.perf_counter()
    for i, (text, wav) in enumerate(chunk_stream):
        times.append(time.perf_counter() - start_time)
        if gen_text != text:
            gen_text = text
            gen_all_text += gen_text
        print(text, wav.shape)
        audios.append(wav.squeeze().cpu().numpy())
        # save_audio_path = os.path.join(ASSETS_DIR, f"image_chatting_segment_stream-{i}-{text}.wav")
        # torchaudio.save(save_audio_path, wav, sample_rate=24000)
        # print(f"Audio saved to {save_audio_path}")
        start_time = time.perf_counter()

    print(f"gen all text: {gen_all_text}")
    save_audio_path = os.path.join(ASSETS_DIR, f"image_chatting_segment_stream.wav")
    sf.write(save_audio_path, np.concatenate(audios), samplerate=24000)
    print(f"All Audio saved to {save_audio_path}")
    info = sf.info(save_audio_path, verbose=True)
    print(
        f"thinker->talker->code2wav chunk streaming first chunk time: {times[0]} s | wav duration: {info.duration} s | cost: {sum(times)} s | RTF: {sum(times)/info.duration}"
    )


def audio_image_chatting_stream():
    import torchaudio
    import soundfile as sf

    messages = [
        {"role": "system", "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}]},
        {
            "role": "user",
            "content": [
                {"type": "audio", "audio": os.path.join(ASSETS_DIR, "image.mp3")},
                {"type": "image", "image": os.path.join(ASSETS_DIR, "03-Confusing-Pictures.jpg")},
            ],
        },
    ]
    # response, audio = inference(
    #    messages, return_audio=True, use_audio_in_video=False, thinker_do_sample=True
    # )
    # print(response[0])
    # save_audio_path = os.path.join(ASSETS_DIR, f"generated_audio_image_chatting_stream.wav")
    # torchaudio.save(save_audio_path, audio, sample_rate=24000)
    # print(f"Audio saved to {save_audio_path}")
    # return

    for _ in range(1):  # warmup and test
        streamer = thinker_talker_inference_stream(messages, use_audio_in_video=False)
        audios = []
        times = []
        start_time = time.perf_counter()
        for i, (texts, audio) in enumerate(streamer):
            if i == 0:
                print(texts[0])
            times.append(time.perf_counter() - start_time)
            audios.append(audio.squeeze().cpu().numpy())
            # save_audio_path = os.path.join(ASSETS_DIR, f"generated_audio_image_chatting_stream_{i}.wav")
            # torchaudio.save(save_audio_path, audio, sample_rate=24000)
            # print(f"Audio saved to {save_audio_path}")
            start_time = time.perf_counter()

        save_audio_path = os.path.join(ASSETS_DIR, f"generated_audio_image_chatting_stream.wav")
        sf.write(save_audio_path, np.concatenate(audios), samplerate=24000)
        info = sf.info(save_audio_path, verbose=True)
        print(
            f"thinker->talker->code2wav streaming first chunk time: {times[0]} s | wav duration: {info.duration} s | cost: {sum(times)} s | RTF: {sum(times)/info.duration}"
        )


def audio_image_chatting_segment_stream():
    import torchaudio
    import soundfile as sf

    messages = [
        {"role": "system", "content": [{"type": "text", "text": SPEECH_SYS_PROMPT}]},
        {
            "role": "user",
            "content": [
                {"type": "audio", "audio": os.path.join(ASSETS_DIR, "image.mp3")},
                {"type": "image", "image": os.path.join(ASSETS_DIR, "03-Confusing-Pictures.jpg")},
            ],
        },
    ]
    thinker_eos_token_ids = [151644, 151645]
    print(thinker_eos_token_ids)
    chunk_stream = generate_stream(
        messages,
        use_audio_in_video=False,
        thinker_max_new_tokens=100,
        thinker_max_tokens_per_step=15,
        thinker_stop_strings_per_step=[".", "。"],
        thinker_eos_token_ids=thinker_eos_token_ids,
    )

    gen_text = ""
    gen_all_text = ""
    audios = []
    times = []
    start_time = time.perf_counter()
    for i, (text, wav) in enumerate(chunk_stream):
        times.append(time.perf_counter() - start_time)
        if gen_text != text:
            gen_text = text
            gen_all_text += gen_text
        print(text, wav.shape)
        audios.append(wav.squeeze().cpu().numpy())
        # save_audio_path = os.path.join(ASSETS_DIR, f"audio_image_chatting_segment_stream-{i}-{text}.wav")
        # torchaudio.save(save_audio_path, wav, sample_rate=24000)
        # print(f"Audio saved to {save_audio_path}")
        start_time = time.perf_counter()

    print(f"gen all text: {gen_all_text}")
    save_audio_path = os.path.join(ASSETS_DIR, f"audio_image_chatting_segment_stream.wav")
    sf.write(save_audio_path, np.concatenate(audios), samplerate=24000)
    print(f"All Audio saved to {save_audio_path}")
    info = sf.info(save_audio_path, verbose=True)
    print(
        f"thinker->talker->code2wav chunk streaming first chunk time: {times[0]} s | wav duration: {info.duration} s | cost: {sum(times)} s | RTF: {sum(times)/info.duration}"
    )


def tokenizer():
    print(processor.tokenizer.pad_token_id)
    print(",", processor.tokenizer.encode(","))
    print(".", processor.tokenizer.encode("."))
    print("?", processor.tokenizer.encode("?"))
    print("，", processor.tokenizer.encode("，"))
    print("。", processor.tokenizer.encode("。"))
    print("？", processor.tokenizer.encode("？"))
    print("！", processor.tokenizer.encode("！"))


def tokenizer_sentences():
    # return processor.tokenizer.encode(";.?!；。？！")
    token_ids = []
    for i in ",;.?，；。？！":
        # for i in ",.":
        token_id = processor.tokenizer.encode(i)
        token_ids.extend(token_id)
    return token_ids


class TokenStreamer(BaseStreamer):
    def __init__(self, skip_prompt: bool = False, timeout=None):
        self.skip_prompt = skip_prompt

        # variables used in the streaming process
        self.token_queue = Queue()
        self.stop_signal = None
        self.next_tokens_are_prompt = True
        self.timeout = timeout

    def put(self, value):
        if len(value.shape) > 1 and value.shape[0] > 1:
            raise ValueError("TextStreamer only supports batch size 1")
        elif len(value.shape) > 1:
            value = value[0]

        if self.skip_prompt and self.next_tokens_are_prompt:
            self.next_tokens_are_prompt = False
            return

        for token in value.tolist():
            self.token_queue.put(token)

    def end(self):
        self.token_queue.put(self.stop_signal)

    def __iter__(self):
        return self

    def __next__(self):
        value = self.token_queue.get(timeout=self.timeout)
        if value == self.stop_signal:
            raise StopIteration()
        else:
            return value


def achatbot_generate():
    import torchaudio
    import soundfile as sf
    from achatbot.core.llm.transformers.manual_vision_voice_qwen import (
        TransformersManualQwen2_5OmniLLM,
    )
    from achatbot.common.session import Session, SessionCtx
    from achatbot.core.llm import LLMEnvInit
    from achatbot.common.logger import Logger

    Logger.init(os.getenv("LOG_LEVEL", "info").upper(), is_file=False, is_console=True)

    session = Session(**SessionCtx("test_client_id", 16000, 2).__dict__)
    args = LLMEnvInit.get_qwen2_5omni_transformers_args()
    args["speaker"] = "Ethan"
    args["lm_attn_impl"] = "flash_attention_2"
    args["warmup_steps"] = 1
    args["warnup_prompt"] = "你叫什么名字？"
    args["is_use_sliding_window_code2wav"] = True
    args["thinker_all_talker_stream"] = False
    args["code2wav_args"]["enable_torch_compile"] = False
    args["code2wav_args"]["enable_torch_compile_first_chunk"] = False
    llm = TransformersManualQwen2_5OmniLLM(**args)

    print("----start generate stream----")

    session.ctx.state["prompt"] = [
        {"type": "text", "text": "请描述一下图片中的内容"},
        {"type": "image", "image": os.path.join(ASSETS_DIR, "03-Confusing-Pictures.jpg")},
        # {"type": "video", "video": ""},
        # {"type": "audio", "audio": ""},
    ]
    kwargs = {
        "use_audio_in_video": False,
        "thinker_top_k": 10,
        "thinker_top_p": 0.9,
        "thinker_temperature": 0.95,
        "thinker_repetition_penalty": 1.1,
        "thinker_min_new_tokens": 1,
        "thinker_max_tokens_per_step": 15,
        "thinker_stop_strings_per_step": [",", ".", "，", "。"],
        "thinker_max_new_tokens": 150,
        "thinker_eos_token_ids": [
            151644,
            151645,
        ],
        "thinker_pad_token_id": 151643,
    }
    chunk_stream = llm.generate(session, **kwargs)
    gen_text = ""
    gen_all_text = ""
    audios = []
    times = []
    start_time = time.perf_counter()
    for i, chunk in enumerate(chunk_stream):
        times.append(time.perf_counter() - start_time)
        text = chunk["text"]
        if gen_text != text:
            gen_text = text
            gen_all_text += gen_text
        if "audio_wav" in chunk:
            wav = chunk["audio_wav"]
            print(text, wav.shape)
            audios.append(wav.squeeze().cpu().numpy())
            # save_audio_path = os.path.join(ASSETS_DIR, f"achatbot_generate_stream-{i}-{text}.wav")
            # torchaudio.save(save_audio_path, wav, sample_rate=24000)
            # print(f"Audio saved to {save_audio_path}")
        else:
            print(text)
        start_time = time.perf_counter()

    print(f"gen all text: {gen_all_text}")
    if len(audios) > 0:
        save_audio_path = os.path.join(ASSETS_DIR, f"achatbot_generate_stream.wav")
        sf.write(save_audio_path, np.concatenate(audios), samplerate=24000)
        print(f"All Audio saved to {save_audio_path}")
        info = sf.info(save_audio_path, verbose=True)
        print(
            f"thinker->talker->code2wav chunk streaming first chunk time: {times[0]} s | wav duration: {info.duration} s | cost: {sum(times)} s | RTF: {sum(times)/info.duration}"
        )


"""
# NOTE: if want to generate speech, need use SPEECH_SYS_PROMPT to generate speech

# asr (audio understanding)
IMAGE_GPU=L4 modal run src/llm/transformers/qwen2_5omni.py --task universal_audio_understanding

# audio to text and speech
IMAGE_GPU=L40s modal run src/llm/transformers/qwen2_5omni.py --task voice_chatting

# vision(video no audio) to text
IMAGE_GPU=L40s modal run src/llm/transformers/qwen2_5omni.py --task video_information_extracting
IMAGE_GPU=L40s modal run src/llm/transformers/qwen2_5omni.py --task screen_recording_interaction

# vision(video with audio) to text and speech
IMAGE_GPU=L40s modal run src/llm/transformers/qwen2_5omni.py --task omni_chatting_for_math
IMAGE_GPU=L40s modal run src/llm/transformers/qwen2_5omni.py --task omni_chatting_for_music

# vision(video with audio) to text and speech with multi rounds chat, but need more GPU memory
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen2_5omni.py --task multi_round_omni_chatting

# batch requests
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen2_5omni.py --task batch_requests

# stream
# text -> text stream
IMAGE_GPU=L4 modal run src/llm/transformers/qwen2_5omni.py --task thinker_stream 
IMAGE_GPU=L4 modal run src/llm/transformers/qwen2_5omni.py --task thinker_chunk_stream 
# image -> text stream
IMAGE_GPU=L4 modal run src/llm/transformers/qwen2_5omni.py --task image_stream
IMAGE_GPU=L4 modal run src/llm/transformers/qwen2_5omni.py --task image_chunk_stream
# audio -> text stream
IMAGE_GPU=L4 modal run src/llm/transformers/qwen2_5omni.py --task asr_stream
IMAGE_GPU=L4 modal run src/llm/transformers/qwen2_5omni.py --task asr_chunk_stream
# video -> text stream
IMAGE_GPU=L40s modal run src/llm/transformers/qwen2_5omni.py --task screen_recording_interaction_stream
IMAGE_GPU=L40s modal run src/llm/transformers/qwen2_5omni.py --task video_information_extracting_stream
IMAGE_GPU=L40s modal run src/llm/transformers/qwen2_5omni.py --task video_information_extracting_chunk_stream
IMAGE_GPU=L40s modal run src/llm/transformers/qwen2_5omni.py --task screen_recording_interaction_chunk_stream

# text -> text + chunk speech stream
IMAGE_GPU=L4 modal run src/llm/transformers/qwen2_5omni.py --task omni_chatting_stream

# text -> chunk text+speech stream
IMAGE_GPU=L40s modal run src/llm/transformers/qwen2_5omni.py --task omni_chatting_segment_stream

# text+image -> text + chunk speech stream
IMAGE_GPU=L4 modal run src/llm/transformers/qwen2_5omni.py --task image_chatting_stream
# text+image -> chunk text+speech stream
IMAGE_GPU=L40s modal run src/llm/transformers/qwen2_5omni.py --task image_chatting_segment_stream

# audio+image -> text + chunk speech stream
IMAGE_GPU=L4 modal run src/llm/transformers/qwen2_5omni.py --task audio_image_chatting_stream
# audio+image -> chunk text+speech stream
IMAGE_GPU=L40s modal run src/llm/transformers/qwen2_5omni.py --task audio_image_chatting_segment_stream

# vision(video with audio) -> text + chunk speech stream
IMAGE_GPU=L40s modal run src/llm/transformers/qwen2_5omni.py --task omni_chatting_for_math_stream
IMAGE_GPU=L40s modal run src/llm/transformers/qwen2_5omni.py --task omni_chatting_for_music_stream

# vision(video with audio) -> chunk text+speech stream
IMAGE_GPU=L40s modal run src/llm/transformers/qwen2_5omni.py --task omni_chatting_for_math_chunk_stream
IMAGE_GPU=L40s modal run src/llm/transformers/qwen2_5omni.py --task omni_chatting_for_music_chunk_stream


# text/vision/audio -> chunk text+speech stream  use sliding window code2wav
ACHATBOT_VERSION=0.0.9.post9 IMAGE_GPU=L40s modal run src/llm/transformers/qwen2_5omni.py --task achatbot_generate

IMAGE_GPU=L4 modal run src/llm/transformers/qwen2_5omni.py --task tokenizer
"""


@app.local_entrypoint()
def main(task: str = "universal_audio_understanding"):
    tasks = {
        "tokenizer": tokenizer,
        "achatbot_generate": achatbot_generate,
        "universal_audio_understanding": universal_audio_understanding,
        "voice_chatting": voice_chatting,
        "video_information_extracting": video_information_extracting,
        "video_information_extracting_stream": video_information_extracting_stream,
        "video_information_extracting_chunk_stream": video_information_extracting_chunk_stream,
        "screen_recording_interaction": screen_recording_interaction,
        "screen_recording_interaction_stream": screen_recording_interaction_stream,
        "screen_recording_interaction_chunk_stream": screen_recording_interaction_chunk_stream,
        "omni_chatting_for_math": omni_chatting_for_math,
        "omni_chatting_for_math_stream": omni_chatting_for_math_stream,
        "omni_chatting_for_math_chunk_stream": omni_chatting_for_math_chunk_stream,
        "omni_chatting_for_music": omni_chatting_for_music,
        "omni_chatting_for_music_stream": omni_chatting_for_music_stream,
        "omni_chatting_for_music_chunk_stream": omni_chatting_for_music_chunk_stream,
        "image_chatting_stream": image_chatting_stream,
        "image_chatting_segment_stream": image_chatting_segment_stream,
        "audio_image_chatting_stream": audio_image_chatting_stream,
        "audio_image_chatting_segment_stream": audio_image_chatting_segment_stream,
        "multi_round_omni_chatting": multi_round_omni_chatting,
        "batch_requests": batch_requests,
        "thinker_stream": thinker_stream,
        "thinker_chunk_stream": thinker_chunk_stream,
        "image_stream": image_stream,
        "image_chunk_stream": image_chunk_stream,
        "asr_stream": asr_stream,
        "asr_chunk_stream": asr_chunk_stream,
        "omni_chatting_stream": omni_chatting_stream,
        "omni_chatting_segment_stream": omni_chatting_segment_stream,
    }
    if task not in tasks:
        raise ValueError(f"task {task} not found")
    print(f"running task {task}")
    run.remote(tasks[task])



# -*- coding: utf-8 -*-
"""PERCENTAGE类
百分数 <=> 中文字符串 方法
中文字符串 <=> 百分数 方法
"""

__author__ = "Zhiyang Zhou <zyzhou@stu.xmu.edu.cn>"
__data__ = "2019-05-06"

from fish_speech.text.chn_text_norm.basic_util import *


class Percentage:
    """
    PERCENTAGE类
    """

    def __init__(self, percentage=None, chntext=None):
        self.percentage = percentage
        self.chntext = chntext

    def chntext2percentage(self):
        return chn2num(self.chntext.strip().strip("百分之")) + "%"

    def percentage2chntext(self):
        return "百分之" + num2chn(self.percentage.strip().strip("%"))


if __name__ == "__main__":

    # 测试程序
    print(Percentage(chntext="百分之五十六点零三").chntext2percentage())
    print(Percentage(percentage="65.3%").percentage2chntext())
